/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.syslog;


import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Calendar;


/**
 *  Parses the raw Syslog messages with output used for further analysis.
 *  Use this instead of Perl parser when dealing with live data stream
 *  to prevent block-buffering of pipes.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SyslogParser.java,v 1.2 2003/09/19 00:38:23 glchen Exp $
 */
public class SyslogParser 
    implements SyslogConstants
{
    protected long _timestamp;
    protected int _year;
    protected String _month;
    protected String _day;
    protected String _time;
    protected String _nshost;
    protected String _apname;
    protected String _msgtype;
    protected String _clienthost;
    protected String _clientmac;
    protected String _action;
    protected String _note;
    
    public SyslogParser(int year)
    {
        _year = year;
    }
    
    public String parse(String line)
    {
        int p1, p2;
        reset();
        checkError(line);
        line = line.replaceAll("[ ]+", " ");
        String[] data = line.split(" ");
        _month = data[0];
        _day = data[1];
        _time = data[2];
        _nshost = data[3];
        _apname = data[4];
        parseMsgType(data[5]);
        if (_msgtype.equals(TAG_SYSLOG))
        {
            //Aug 12 07:40:59 ns1.dartmouth.edu last message repeated 2 times
            _action = TAG_REPEAT_LAST_MSG;
            _note = data[7];
        }
        else if (data[6].equals("Station"))
        {
            //Aug 12 04:10:00 ns1.dartmouth.edu sachem-42-wb (Warning): 
            //Station [sachem-16-wb]00407824d218 Failed Association, status "Too Many Stations in BSS"
            // or
            //Aug 12 04:10:00 ns1.dartmouth.edu sachem-104-wb (Info): 
            //Station [sachem-16-wb]00409782d218 Authenticated
            p1 = line.indexOf("]");
            if (p1 != -1)
            {
                p2 = line.indexOf(" ", p1);
                p1 = line.indexOf("[");
                parseClient(line.substring(p1, p2));
            }
            else
            {
                parseClient(data[7]);
            }
            p1 = line.indexOf(_clientmac);
            p2 = line.indexOf(",", p1);
            if (p2 == -1)
            {
                _action = line.substring(p1 + 12);
            }
            else
            {
                _action = line.substring(p1+12, p2);
                parseNote(line);                
            }
        }
        else if (line.matches("^.*[0-9a-f]{12}.*$"))
        {
            //Aug 12 11:57:15 ns1.dartmouth.edu berry17-ap (Info): 
            //Deauthenticating 0030782b5c1b, reason "Must Authenticate Before Associating"
            p1 = line.indexOf("): ");
            p2 = line.indexOf(",", p1);
            String c = null;
            if (p2 != -1)
            {
                c = line.substring(p1+3, p2);
                parseNote(line);
            }
            else
            {
                c = line.substring(p1+3);
            }
            p1 = c.indexOf("[");
            if (p1 != -1)
            {
                _action = c.substring(0, p1);
                parseClient(c.substring(p1));
            }
            else
            {
                p1 = c.lastIndexOf(" ");
                _action = c.substring(0, p1);
                parseClient(c.substring(p1+1));
            }
        }
        else if (line.indexOf("):") != -1)
        {
            p1 = line.indexOf("): ");
            _note = line.substring(p1+3);
        }
        else
        {
            throw new IllegalArgumentException(
                "Unparsable message {" + line + "}");
        }
        return compose();
    }

    private void reset()
    {
        _month = null;
        _day = null;
        _time = null;
        _nshost = null;
        _apname = null;
        _msgtype = null;
        _clienthost = null;
        _clientmac = null;
        _action = null;
        _note = null;
    }

    private void checkError(String line)
    {
        if (line.matches("3\\^M\\s,\\?M$"))
        {
            throw new IllegalArgumentException("Error {" + line + "}");
        }
    }
    
    private void parseMsgType(String msgtype)
    {
        if (msgtype.equals("message"))
        {
            _msgtype = TAG_SYSLOG;
        }
        else if (msgtype.equals("(Warning):"))
        {
            _msgtype = TAG_WARNING;
        }
        else if (msgtype.equals("(Info):"))
        {
            _msgtype = TAG_INFO;
        }
        else if (msgtype.equals("(FATAL):"))
        {
            _msgtype = TAG_FATAL;
        }
        else
        {
            throw new IllegalArgumentException(
                "Unknown message type {" + msgtype + "}");
        }
    }

    private void parseClient(String client)
    {
        if (client.startsWith("["))
        {
            int p = client.indexOf("]");
            _clienthost = client.substring(1, p);
            _clientmac = client.substring(p+1);
        }
        else
        {
            _clientmac = client;
        }
        if (_clienthost != null && _clienthost.startsWith("?"))
        {
            _clienthost = _clienthost.substring(1);
        }
    }

    private void parseNote(String line)
    {
        int p1 = line.indexOf("\"");
        int p2 = line.indexOf("\"", p1 + 1);
        if (p1 != -1 && p2 != -1)
        {
            _note = line.substring(p1+1, p2);
        }
        else
        {
            p1 = line.indexOf(",");
            if (p1 != -1)
            {
                _note = line.substring(p1+1);
            }
        }
    }

    private String compose()
    {
        StringBuffer ret = new StringBuffer();
        ret.append(getTimeStamp() + "\t");
        ret.append(_year + "\t");
        ret.append(_month + "\t");
        ret.append(_day + "\t");
        ret.append(_time + "\t");
        ret.append(_msgtype + "\t");
        ret.append(clean(_apname) + "\t");
        ret.append(clean(_clienthost) + "\t");
        ret.append(clean(_clientmac) + "\t");
        ret.append(clean(_action) + "\t");
        ret.append(clean(_note) + "\t");
        return ret.toString();
    }

    private String clean(String field)
    {
        if (field == null)
        {
            return "[]";
        }
        field = field.trim().replaceAll(" ", "#");
        return field;
    }

    private String getTimeStamp()
    {
        Calendar cal = Calendar.getInstance();
        cal.clear();
        cal.set(Calendar.YEAR, _year);
        if (_month.equals("Jan"))
            cal.set(Calendar.MONTH, 0);
        else if (_month.equals("Feb"))
            cal.set(Calendar.MONTH, 1);
        else if (_month.equals("Mar"))
            cal.set(Calendar.MONTH, 2);
        else if (_month.equals("Apr"))
            cal.set(Calendar.MONTH, 3);
        else if (_month.equals("May"))
            cal.set(Calendar.MONTH, 4);
        else if (_month.equals("Jun"))
            cal.set(Calendar.MONTH, 5);
        else if (_month.equals("Jul"))
            cal.set(Calendar.MONTH, 6);
        else if (_month.equals("Aug"))
            cal.set(Calendar.MONTH, 7);
        else if (_month.equals("Sep"))
            cal.set(Calendar.MONTH, 8);
        else if (_month.equals("Oct"))
            cal.set(Calendar.MONTH, 9);
        else if (_month.equals("Nov"))
            cal.set(Calendar.MONTH, 10);
        else if (_month.equals("Dec"))
            cal.set(Calendar.MONTH, 11);
        String[] t = _time.split(":");
        cal.set(Calendar.HOUR_OF_DAY, Integer.parseInt(t[0]));
        cal.set(Calendar.MINUTE, Integer.parseInt(t[1]));
        cal.set(Calendar.SECOND, Integer.parseInt(t[2]));
        return Long.toString(cal.getTimeInMillis()).substring(0, 10);
    }

    //=======================================================================
    // Test.
    //=======================================================================
    
    public static void main(String[] args)
        throws Exception
    {
        SyslogParser parser = new SyslogParser(2003);
        BufferedReader in = 
            new BufferedReader(new InputStreamReader(System.in));
        String line = null;
        while ((line = in.readLine()) != null)
        {
            try
            {
                System.out.println(parser.parse(line));
//                System.out.flush();
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                System.err.println(line);
//                System.out.flush();
                break;
            }
        }
    }
}
