/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */
 
package campus.sysnort;


import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.File;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import javax.xml.parsers.*;
import solar.api.ParsingException;


/**
 *  An XML parser that parses the Sysnort configuration file into a set
 *  of triggers.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: XmlConfigParser.java,v 1.2 2003/08/22 00:54:27 glchen Exp $
 */
public class XmlConfigParser 
	extends DefaultHandler
{
    // constants
    public static final String SYSNORT      = "sysnort";
	public static final String TRIGGER      = "trigger";
	public static final String NAME         = "name";
	public static final String CLASSNAME    = "classname";
	public static final String CODEBASE     = "codebase";
	public static final String ARGUMENT     = "argument";
	public static final String VALUE        = "value";
    

    // parser factory
	private static final SAXParserFactory _factory = 
		SAXParserFactory.newInstance();

    // current parsing state
    protected SysnortConfig _config;
    protected TriggerConfig _trigger;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public XmlConfigParser()
    {
    }


    //=======================================================================
    // Implementation of solar.service.dag.SpecParser interface.
    //=======================================================================

    public SysnortConfig parse(File file)
        throws ParsingException
    {
		try
		{
			FileInputStream in = new FileInputStream(file);
			SysnortConfig ret = parse(in);
			in.close();
			return ret;
		}
		catch (IOException ioe)
		{
			throw new ParsingException(ioe);
		}
    }

    public SysnortConfig parse(InputStream in)
        throws ParsingException
    {
		return parse(new InputSource (in));
    }

    public SysnortConfig parse(Reader reader)
        throws ParsingException
    {
		return parse(new InputSource(reader));
    }

    public SysnortConfig parse(String xml)
        throws ParsingException
    {
		StringReader reader = new StringReader(xml);
		return parse(new InputSource(reader));
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

	public static SysnortConfig parse(InputSource in)
		throws ParsingException
	{
		try
		{
			SAXParser sax = _factory.newSAXParser();
			XmlConfigParser parser = new XmlConfigParser();
			sax.parse(in, parser);
			return parser.getSysnortConfig();
		}
        catch (SAXException saxe)
        {
			throw new ParsingException(saxe);
        }
		catch (ParserConfigurationException pce)
		{
			throw new ParsingException(pce);
		}
		catch (IOException ioe)
		{
			throw new ParsingException(ioe);
		}
	}


    //=======================================================================
    // Parsing methods.
    //=======================================================================

    /** Return recent parsed graph spec. */
    public SysnortConfig getSysnortConfig()
    {
        return _config;
    }

	public void startElement (String uri, String localName, 
        String name, Attributes attrs)
		throws SAXException
	{
		switch (name.charAt (0))
		{
		case 's':
            assertEquals(name, SYSNORT);
            startSysnort(attrs);
			break;
		case 't':
            assertEquals(name, TRIGGER);
            startTrigger(attrs);
			break;
		case 'a':
            assertEquals(name, ARGUMENT);
            startArgument(attrs);
			break;
		default:
			assertEquals(name, "");	// will throw proper exception
		}
    }

	public void endElement(String uri, String localName, String name)
		throws SAXException
	{
		switch (name.charAt (0))
		{
		case 't':
            assertEquals(name, TRIGGER);
            endTrigger();
			break;
		}
	}

    private void startSysnort(Attributes attrs)
    {
        _config = new SysnortConfig();
    }

    private void startTrigger(Attributes attrs)
        throws SAXException
    {
        String name = attrs.getValue(NAME);
        if (name == null)
            throw new SAXException(
                "Trigger must have a symbolic name.");
        if (_config.getTrigger(name) != null)
            throw new SAXException(
                "Trigger " + name + " has already been defined.");
            
        String clazz = attrs.getValue(CLASSNAME);
        String codebase = attrs.getValue(CODEBASE);
        _trigger = new TriggerConfig();
        _trigger.setName(name);
        _trigger.setClassname(clazz);
        _trigger.setCodebase(codebase);
    }

    private void endTrigger()
    {
        _config.putTrigger(_trigger.getName(), _trigger);
        _trigger = null;
    }

    private void startArgument(Attributes attrs)
    {
        String name = attrs.getValue(NAME);
        String value = attrs.getValue(VALUE);
        _trigger.addArgument(name, value);
    }

	private void assertEquals (String s1, String s2)
		throws SAXException
	{
		if (!(s1.equals (s2)))
			throw new SAXException("Illegal tag: " + s1);
	}


    //=======================================================================
    // Test.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        XmlConfigParser p = new XmlConfigParser();
        p.parse(new File(args[0]));
    }
}
