/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package rice.pastry.socket;


import java.util.*;
import java.net.*;
import java.io.*;
import java.nio.channels.*;
import java.nio.*;

import rice.pastry.*;
import rice.pastry.messaging.*;
import rice.pastry.security.*;
import rice.pastry.standard.*;
import rice.pastry.routing.*;
import rice.pastry.leafset.*;
import rice.pastry.dist.*;
import rice.pastry.wire.messaging.socket.*;

import solar.api.*;
import solar.impl.*;
import solar.service.transport.*;


/**
 *  A direct mimic to WirePastryNodeFactory class.
 *  An abstraction of the nodeId factory for distributed nodes. In order 
 *  to obtain a nodeId factory, a client should use the getFactory method, 
 *  passing in either PROTOCOL_TCP or PROTOCOL_UDP as the protocol, and the 
 *  port number the factory should use.
 *
 *  @author  Guanling Chen
 *  @author  Alan Mislove
 *  @version $Id: SocketPastryNodeFactory.java,v 1.5 2003/10/05 18:50:53 glchen Exp $
 */
public class SocketPastryNodeFactory 
    extends DistPastryNodeFactory
    implements TransportListener
{
    // choices of protocols
    public static int PROTOCOL_TCP = 0;
    public static int PROTOCOL_UDP = 1;

    public static final int rtMax = 1;
    public static final int lSetSize = 24;
    public static final int maxOpenSockets = 5;
    public static final int leafSetMaintFreq = 1*60;
    public static final int routeSetMaintFreq = 15*60;

    protected NodeIdFactory _nidFactory;
    protected int _protocol;
    protected int _port;

    protected boolean _serialize;
    protected Transport _transport;
    protected EventFactory _factory;
    protected InetSocketAddress _laddr;
    protected NodeIdResponseMessage _response;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public SocketPastryNodeFactory(boolean serialize, 
        NodeIdFactory nf, int proto, int startPort) 
    {
        _serialize = serialize;
        _nidFactory = nf;
        _port = startPort;
        _protocol = proto;

        // create a new TCP transport with random port until success
        int tport = 2000;
        Random ran = new Random(System.currentTimeMillis());
        tport += ran.nextInt(100);
        while (true)
        {
            try
            {
                if (_protocol == PROTOCOL_TCP)
                {
                    _transport = new TcpTransport(tport);
                }
                else
                {
                    _transport = new UdpTransport(tport);
                }
                break;
            }
            catch (Exception ex)
            {
                tport += ran.nextInt(100);
            }
        }
        _factory = new EventFactory();
        _laddr = getAddress(tport);
        _transport.addTransportListener(this);
    }

    //=======================================================================
    // Implementation of TransportListener interface.
    //=======================================================================

//    public void dataReceived(byte[] data)
    public void dataReceived(Adu adu, Header header)
    {
        try
        {
//            Event evt = (Event) _factory.parse(data);
            Event evt = (Event) adu;
            debug("SocketPastryNodeFactory: " + evt);
            _response = (NodeIdResponseMessage) evt.getValue("message");
            synchronized (_transport)
            {
                _transport.notifyAll();
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }

    public void connectionOpened(solar.api.Address peer)
    {
        
    }

    public void connectionClosed(solar.api.Address peer)
    {
        
    }


    //=======================================================================
    // Implementation of PastryNodeFactory interface.
    //=======================================================================

    /**
     * Method which contructs a node handle (using the wire protocol) for the
     * node at address NodeHandle.
     *
     * @param address The address of the remote node.
     * @return A NodeHandle cooresponding to that address
     */
    public NodeHandle generateNodeHandle(InetSocketAddress address) 
    {
        // send nodeId request to remote node, wait for response
        // allocate enought bytes to read a node handle
        SocketNodeHandle handle = null;

        System.out.println("Socket: Contacting bootstrap node " + address);

        try {
            _response = null;
            // send NodeIdRequest event
            List attr = new Vector();
            attr.add(new Attribute("address", _laddr));
            attr.add(new Attribute("message", new NodeIdRequestMessage()));
            _transport.send(
                EventFactory.getSerializedEvent(attr).toBytes(), 
                new solar.service.transport.SocketAddress(address));

            // wait till reply comes in or time out
            synchronized (_transport)
            {
                _transport.wait(5000);
            }
            if (_response == null)
            {
                throw new Exception("NodeIdRequest timeout.");
            }
            debug("Read " + _response + " from remote node...");
            handle = new SocketNodeHandle(_protocol, address, _response.getNodeId());
            return handle;
        }
        catch (Exception e) 
        {
            System.out.println("Error connecting to address " + address + ": " + e);
            System.out.println("Couldn't find a bootstrap node, starting a new ring...");
            return null;
        }
    }


    /**
     * Method which creates a Pastry node from the next port
     * with a randomly generated NodeId.
     *
     * @param bootstrap Node handle to bootstrap from.
     * @return A node with a random ID and next port number.
     */
    public PastryNode newNode(NodeHandle bootstrap) 
    {
        return newNode(bootstrap, _nidFactory.generateNodeId());
    }
  
    /**
     * Method which creates a Pastry node from the next port
     * with a randomly generated NodeId.
     *
     * @param bootstrap Node handle to bootstrap from.
     * @return A node with a random ID and next port number.
     */
    public PastryNode newNode(final NodeHandle bootstrap, NodeId nodeId) 
    {
        final SocketPastryNode pn = new SocketPastryNode(nodeId);

        Transport transport = null;
        InetSocketAddress address = null;
        synchronized (this) 
        {
            try
            {
                if (_protocol == PROTOCOL_TCP)
                {
                    transport = (_serialize) ? 
                        new TcpObjectTransport(_port) :
                        new TcpTransport(_port);
                }
                else
                {
                    transport = (_serialize) ? 
                        new UdpObjectTransport(_port) :
                        new UdpTransport(_port);
                }
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                return null;
            }
            address = getAddress(_port);
            _port++;
        }
      
        SocketNodeHandle localhandle = new SocketNodeHandle(_protocol, address, nodeId);
        transport.addTransportListener(localhandle);
        SocketNodeHandlePool pool = new SocketNodeHandlePool(pn);
        SocketPastrySecurityManager secureMan = 
            new SocketPastrySecurityManager(localhandle, pool);
            
        // didn't touch following code
        MessageDispatch msgDisp = new MessageDispatch();
        RoutingTable routeTable = new RoutingTable(localhandle, rtMax);
        LeafSet leafSet = new LeafSet(localhandle, lSetSize);
        StandardRouter router =
            new StandardRouter(localhandle, routeTable, leafSet, secureMan);
        StandardLeafSetProtocol lsProtocol =
            new StandardLeafSetProtocol(pn, localhandle, secureMan, leafSet, routeTable);
        StandardRouteSetProtocol rsProtocol =
            new StandardRouteSetProtocol(localhandle, secureMan, routeTable);
        StandardJoinProtocol jProtocol =
            new StandardJoinProtocol(pn, localhandle, secureMan, routeTable, leafSet);
        msgDisp.registerReceiver(router.getAddress(), router);
        msgDisp.registerReceiver(lsProtocol.getAddress(), lsProtocol);
        msgDisp.registerReceiver(rsProtocol.getAddress(), rsProtocol);
        msgDisp.registerReceiver(jProtocol.getAddress(), jProtocol);
        pn.setElements(localhandle, secureMan, msgDisp, leafSet, routeTable);
        
        pn.setSocketElements(address, 
            transport, pool, leafSetMaintFreq, routeSetMaintFreq);
        
        // didn't touch following code
        secureMan.setLocalPastryNode(pn);
        pool.coalesce(localhandle);
        localhandle.setLocalNode(pn);
        if (bootstrap != null)
        {
            bootstrap.setLocalNode(pn);
            pn.doneNode(bootstrap);
        }
        return pn;
    }

    /**
     *  Method which constructs an InetSocketAddres for the local host
     *  with the specifed port number.
     *
     *  @param portNumber The port number to create the address at.
     *  @return An InetSocketAddress at the localhost with port portNumber.
     */
    private InetSocketAddress getAddress(int portNumber) 
    {
        InetSocketAddress result = null;

        try {
        result = new InetSocketAddress(InetAddress.getLocalHost(), portNumber);
        } catch (UnknownHostException e) {
        System.out.println("PANIC: Unknown host in getAddress. " + e);
        }

        return result;
    }

    private void debug(String s) 
    {
        if (Log.ifp(8))
            System.out.println(" (F): " + s);
    }
    
  

    /**
    * Static method which is designed to be used by clients needing a
    * distrubuted pastry node factory.  The protocol should be one of
    * PROTOCOL_TCP or PROTOCOL_UDP.
    *
    * @param protocol The protocol to use (PROTOCOL_TCP or PROTOCOL_UDP)
    * @param port The starting port.
    * @return A DistPastryNodeFactory using the given protocol and port.
    * @throws IllegalArgumentException If protocol is an unsupported port.
    */
    public static DistPastryNodeFactory getFactory(
        boolean serialize,
        NodeIdFactory nf, int protocol, int port) 
    {
        if (protocol == PROTOCOL_TCP)
        {
            return new SocketPastryNodeFactory(serialize, nf, protocol, port);
        }
        else if (protocol == PROTOCOL_UDP)
        {
            return new SocketPastryNodeFactory(serialize, nf, protocol, port);
        }
        else
        {
            throw new IllegalArgumentException(
                "Unsupported Protocol " + protocol);
        }
    }
}
