/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.impl;


import java.util.List;
import java.util.Vector;
import java.util.Stack;
import java.util.Iterator;

import solar.api.Attribute;
import solar.api.AttributeRecord;
import solar.api.ParsingException;


/**
 *  Plain attribute record encodes string attributes in square brackets.
 *  The attribute values have to be strings.  The brackets "[" and "]"
 *  thus shall not be used in attribute keys and values.
 *  <P>
 *  For example: [ a = [ b = 12 ], c = 33, d = [ e = [ f = 44 ] ] ]
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PlainAttributeRecord.java,v 1.6 2003/10/05 15:24:39 glchen Exp $
 */
public class PlainAttributeRecord
    extends AbstractAttributeRecord
{
    // constants for the parser's state machine
    private static final int KEY = 1;
    private static final int VALUE = 2;
    private static final int DELIMIT = 3;
        

    //=======================================================================
    // Constructors.
    //=======================================================================

    /** Constructs an empty attribute record. */
    protected PlainAttributeRecord()
    {
        _attributes = new Vector();
    }

    
    /** Constructs an attribute record by parsing an encoded string. */
    public PlainAttributeRecord(String rec)
        throws ParsingException
    {
        _attributes = parsePlainAttributeRecord(rec);
    }


    /** Constructs an atribute record with a list of attributes. */
    public PlainAttributeRecord(List aList)
    {
        _attributes = new Vector(aList);
    }

    
    /** Constructs an atribute record by copying from another one. */
    public PlainAttributeRecord(AttributeRecord aRec)
    {
        _attributes = new Vector(aRec.getAttributes());
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    /** Parse an attribute record encoded in a string. */
    private List parsePlainAttributeRecord(String rec)
        throws ParsingException
    {
        // keys and corresponding records in the stacks represents 
        // unfinished parsing attributes.
        Stack keys = new Stack();
        Stack records = new Stack();
        PlainAttributeRecord curRec = null;

        // following hand-coded parser goes through the string as
        // a sequence of chars.
        String key = null;
        String val = null;
        int pos = 0;
        int expect = VALUE;     // represents parsing state
        int len = rec.length();
        for (int i=0; i<len; i++)
        {
            char ch = rec.charAt(i);
            switch (ch)
            {
            case '[':
                if (expect == KEY) throw new ParsingException("Pasring error.");

                // value of current attribute is a new record
                key = (key == null) ? "root" : key;
                keys.push(key);
                curRec = new PlainAttributeRecord();
                records.push(curRec);
                pos = i + 1;
                expect = KEY;
                break;
            case ']':
                if (records.empty()) throw new ParsingException("Parsing error.");
                if (expect == KEY) throw new ParsingException("Pasring error.");
                
                if (expect == VALUE)    // key = val ]
                {
                    if (key == null) throw new ParsingException("Parsing error.");
                    // make a new attribute and add to current record
                    val = rec.substring(pos, i).trim();
                    curRec._attributes.add(new Attribute(key, val));
                }
                //System.out.println("poping up " + keys.peek());
                // a record finishes here, poping up stacks for one-level
                Attribute attr = new Attribute(
                    (String) keys.pop(), records.pop());
                if (!records.empty())
                {
                    curRec = (PlainAttributeRecord) records.peek();
                    curRec._attributes.add(attr);
                }
                key = null;
                val = null;
                pos = i + 1;
                expect = DELIMIT; // looking for , or more ]
                break;
            case ',':
                if (expect == KEY) throw new ParsingException("Pasring error.");
                if (expect == VALUE) // key = val ,
                {
                    if (key == null) throw new ParsingException("Pasring error.");
                    val = rec.substring(pos, i).trim();
                    curRec._attributes.add(new Attribute(key, val));
                }
                else    // ] ,
                {
                    // nothing to do
                }
                key = null;
                val = null;
                pos = i + 1;
                expect = KEY;
                break;
            case '=':
                if (expect != KEY) throw new ParsingException("Pasring error.");
                if (val != null) throw new ParsingException("Pasring error.");
                
                key = rec.substring(pos, i).trim();
                val = null;
                pos = i + 1;
                expect = VALUE;
                break;
            default:
            }
        }

        if (!records.empty())
            throw new ParsingException("Pasring error.");

        return curRec._attributes;
    }


    //=======================================================================
    // Implementation of solar.api.AttributeRecord interface.
    //=======================================================================

    public boolean equals(Object obj)
    {
        if (obj == null || !(obj instanceof PlainAttributeRecord))
            return false;

        PlainAttributeRecord other = (PlainAttributeRecord) obj;
        return _attributes.equals(other._attributes);
    }


    //=======================================================================
    // Implementation of solar.api.Adu interface.
    //=======================================================================
   
    public String toString()
    {
        // converts to brackets enclosed format, separating attributes
        // with commas.
        String ret = "[";
        Iterator itr = _attributes.iterator();
        while (itr.hasNext())
        {
            Attribute attr = (Attribute) itr.next();
            ret += attr.toString();
            ret += (itr.hasNext()) ? "," : "";
        }
        ret += "]";
        return ret;
    }

    public byte[] toBytes()
    {
        return toString().getBytes();
    }
}
