/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.module.cache;


import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.TimerTask;


/**
 *  A base implementation of cache interface.  Once an object is cached,
 *  it cannot be refreshed which means it will be stale at pre-determined
 *  expiration time.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: BaseCache.java,v 1.1 2003/10/04 01:49:26 glchen Exp $
 */
public class BaseCache 
    extends TimerTask
    implements Cache, CacheConstants
{
    protected int _capacity;
    protected long _ttl;
    protected int _size;
    protected List _listeners = new Vector();
    protected Hashtable _cache = new Hashtable();

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public BaseCache()
    {
        _capacity = CACHE_CAPACITY_DEFAULT;
        _ttl = CACHE_TTL_DEFAULT;
    }

    public BaseCache(int capacity, long ttl)
    {
        _capacity = capacity;
        _ttl = ttl;
    }


    //=======================================================================
    // Implementation of solar.service.cache.Cache interface.
    //=======================================================================

    public void addListener(CacheListener lis)
    {
        _listeners.add(lis);
    }

    public void removeListener(CacheListener lis)
    {
        _listeners.remove(lis);
    }
    
    /**
     *  If the objectId already exists in the cache, this add operation does
     *  override the previous cached object with the one supplied in parameter.
     *  But the object is not refreshed meaning expiration time will not 
     *  be changed.
     */
    public synchronized void addObject(Object objectId, Object obj)
        throws CacheException
    {
        CachedObject co;
        if (_cache.get(objectId) == null)
        {
            if ((_capacity != CACHE_CAPACITY_INFINIT) && (_size >= _capacity))
                throw new CacheException("Cannot add more objects to cache.");

            co = new CachedObject(objectId, obj);
            co.setCreated();
            co.setExpires(_ttl);
            _size ++;
        }
        else
        {
            co = (CachedObject) _cache.get(objectId);
            co.setObject(obj);
        }
        _cache.put(objectId, co);
    }

    public synchronized Object getObject(Object objectId)
    {
        Object ret = null;
        CachedObject co = (CachedObject) _cache.get(objectId);
        if (co != null)
            ret = co.getObject();
        return ret;
    }

    public synchronized CachedObject getCachedObject(Object objectId)
    {
        return (CachedObject) _cache.get(objectId);
    }

    public synchronized List getObjects()
    {
        List ret = new Vector();
        Iterator itr = _cache.values().iterator();
        while (itr.hasNext())
        {
            CachedObject co = (CachedObject) itr.next();
            ret.add(co.getObject());
        }
        return ret;
    }

    public synchronized List getCachedObjects()
    {
        List ret = new Vector();
        Iterator itr = _cache.values().iterator();
        while (itr.hasNext())
        {
            ret.add(itr.next());
        }
        return ret;        
    }

    public synchronized List getObjectIds()
    {
        List ret = new Vector(_cache.keySet());
        return ret;        
    }

    public synchronized void removeObject(Object objectId)
    {
        _cache.remove(objectId);
        _size --;
    }

    public synchronized int getCapacity()
    {
        return _capacity;
    }

    public synchronized int getSize()
    {
        return _cache.size();
    }

    public synchronized void flush()
    {
        _cache.clear();
        _size = 0;
    }


    //=======================================================================
    // Implementation of java.util.TimerTask interface.
    //=======================================================================

    public synchronized void run()
    {
        Iterator itr = new Vector(_cache.keySet()).iterator();
        while (itr.hasNext())
        {
            Object key = itr.next();
            CachedObject co = (CachedObject) _cache.get(key);
            if (co.isStale())
            {
                removeObject(key);
                notifyObjectExpired(co);
            }
        }
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void notifyObjectExpired(CachedObject co)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            CacheListener lis = (CacheListener) itr.next();
            lis.notifyObjectExpired(co);
        }
    }
}
