/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service;


import serp.util.Options;

import solar.api.Event;
import solar.api.Address;
import solar.api.Request;
import solar.service.dispatch.DispatchUnit;


/**
 *  A base service implementation.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: BaseService.java,v 1.23 2003/10/05 18:50:54 glchen Exp $
 */
public class BaseService 
    implements Service
{
    protected ServiceManager _manager;

    protected String _name;
    protected boolean _inited = false;


    //=======================================================================
    //  Public interface methods.
    //=======================================================================

    public void setInited(boolean flag)
    {
        _inited = flag;
    }


    //=======================================================================
    //  Private helper methods.
    //=======================================================================

    /** Check whether the service has been initialized. */
    protected void assertInited()
    {
        if (_inited == false)
        {
            throw new ServiceRuntimeException(
                "Service '" + getName() + "' has not been initialized.");
        }
    }

    /**
     *  This method is provided so service can specify a different name key
     *  for other services used by it. For instance:
     *  
     *  service.transport1.classname = solar.service.transport.UdpTransportService
     *  service.transport2.classname = solar.service.transport.TcpTransportService
     *  service.myserver.transport = transport1
     *  service.yourserver.transport = transport2
     */
    protected Service getDependentService(String nameKey, String defaultName)
        throws ServiceException
    {
        if (_manager == null)
        {
            throw new ServiceException("Failed to find service manager.");
        }
        Options config = _manager.getServiceConfig(getName());
        String serviceName = config.getProperty(nameKey, defaultName);
        return _manager.getService(serviceName);
    }
    
    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void setServiceManager(ServiceManager manager)
    {
        _manager = manager;
    }

    public ServiceManager getServiceManager()
    {
        return _manager;
    }

    public void setName(String name)
    {
        _name = name;
    }

    public String getName()
    {
        return _name;
    }

    public void init()
        throws ServiceException
    {
    }

    public void initDependents()
        throws ServiceException
    {
        _inited = true;
    }

    public void init(Object param)
        throws ServiceException
    {
        _inited = true;
    }
    
    public boolean hasInited()
    {
        return _inited;
    }

    public void start()
    {
        // default does nothing
    }
    
    public void stop()
    {
        _inited = false;
    }

    public void handleDispatchUnit(DispatchUnit dunit)
        throws Exception
    {
        // default to call handleEvent and ignore other info
        Event evt = (Event) dunit.getAdu();
        handleEvent(evt, dunit.getSourceAddress());
    }

    public Object handleRequest(Request req)
        throws Exception
    {
        // default does nothing
        return null;
    }

    public void handleEvent(Event evt, Address source)
        throws Exception
    {
        handleEvent(evt);
    }
    
    public void handleEvent(Event evt)
        throws Exception
    {
        // default does nothing
    }
}
