/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.ArrayList;


/**
 *  A base implementation of service manager.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: BaseServiceManager.java,v 1.11 2003/10/02 06:14:28 glchen Exp $
 */
public class BaseServiceManager 
    implements ServiceManager
{
    /** A prefix for service properties configuration file. */
    public static final String SERVICE_PREFIX = "service.";

    /** A service property determining its implementing class name. */
    public static final String CLASSNAME_SUFFIX = ".classname";

    /** Logging facility. */
    protected static Logger _log = Logger.getLogger(BaseServiceManager.class);

    /** Configurations used for all the services. */
    protected Options _config;

    /** Mapping of service names to class names. */
    protected Hashtable _mapping = new Hashtable();

    /** A repository of service instances. */
    protected Hashtable _services = new Hashtable();

    /** A repository of application-specific service objects. */
    protected Hashtable _serviceObjects = new Hashtable();


    //=======================================================================
    //  Constructors.
    //=======================================================================
    
    protected BaseServiceManager()
    {
    }


    //=======================================================================
    //  Implementation of solar.service.ServiceManager interface.
    //=======================================================================

    public void init(Options opts)
        throws ServiceException
    {
        if (_config != null)
        {
            throw new ServiceException(
                "ServiceManager has already been initialized before.");
        }
        _config = opts;
        
        initMapping();
        initServices();
    }
        
    public boolean hasRegistered(String name)
    {
        return (_services.get(name) != null);
    }
    
    public Service getService(String name)
        throws ServiceException
    {
        Service service = getServiceInstance(name);
        if (!service.hasInited())
        {
            synchronized (service.getClass())
            {
                if (!service.hasInited())
                {
//                    service.init();
                    service.initDependents();
//                    _log.debug("Finished initializing service (late): " + name);
                }
            }
        }
        if (!service.hasInited())
        {
            // this exception will be caught & rethrown by this very method.
            // hasInited() returning false indicates some initialization issue,
            // which in turn prevents the InitableBroker from passing a
            // reference to a working instance of the initable to the client.
            throw new ServiceException(
                    "Failed to initialize service " + name);
        }
        return service;
    }

    public Options getServiceConfig(String name)
        throws ServiceException
    {
        Options ret = new Options();
        for (Iterator keys = _config.keySet().iterator(); keys.hasNext();)
        {
            String key = (String) keys.next();
            if (key.startsWith(SERVICE_PREFIX + name))
            {
                int pos = key.lastIndexOf(".");
                ret.setProperty(
                    key.substring(pos + 1), _config.getProperty(key));
            }
        }
        if (ret.size() == 0)
            throw new ServiceException(
                "Requested service [" + name + "] not found.");
        return ret;
    }

    public void setServiceObject(String name, Object obj)
    {
        _serviceObjects.put(name, obj);
    }

    public Object getServiceObject(String name)
    {
        return _serviceObjects.get(name);
    }

    public void stopService(String name)
    {
        _log.info("Shutting down service: " + name);
        try
        {
            Service service = getServiceInstance(name);
            if (service != null && service.hasInited())
            {
                service.stop();
                if (service.hasInited() && service instanceof BaseService)
                {
                    // BaseService::stop() does this by default,
                    // but could've been overriden poorly.
                    ((BaseService) service).setInited(false);
                }
            }
        }
        catch (Exception ex)
        {
            // Assuming harmless -- log the error and continue.
            _log.error("Shutdown of a nonexistent service: " + name, ex);
        }
    }

    public void stopServices()
    {
        _log.info("Shutting down all services!");
        for (Iterator names = _mapping.keySet().iterator(); names.hasNext();)
        {
            stopService((String) names.next());
        }

    }


    //=======================================================================
    //  Private helper methods.
    //=======================================================================

    /**
     * Creates a mapping between service names and class names.
     *
     * The mapping is built according to settings present in
     * configuration file.  The entries should have the
     * following form:
     *
     * <pre>
     * service.MyService.classname=com.mycompany.MyServiceImpl
     * service.MyOtherService.classname=com.mycompany.MyOtherServiceImpl
     * </pre>
     */
    private void initMapping()
    {
        for (Iterator keys = _config.keySet().iterator(); keys.hasNext();)
        {
            // a key in property file is a string separated by "."
            String key = (String) keys.next();
            String[] keyParts = key.split("\\.");

            // a service is identified by this format:
            // service.xxxxxx.classname = something
            if ((keyParts.length == 3)
                    && (keyParts[0] + ".").equals(SERVICE_PREFIX)
                    && ("." + keyParts[2]).equals(CLASSNAME_SUFFIX))
            {
                String serviceName = keyParts[1];
                if (!_mapping.containsKey(serviceName))
                {
                    // put the service name and its classname into 
                    // the mapping table
                    _mapping.put(serviceName, _config.getProperty(key));
                }
            }
        }
    }

    /** Performs early initialization of all needed services. */
    private void initServices()
        throws ServiceException
    {
        for (Iterator names = _mapping.keySet().iterator(); names.hasNext();)
        {
            try
            {
                String name = (String) names.next();
                Service svc = getServiceInstance(name);
                svc.init();
//                _log.debug("Finished initializing service (early): " + name);
            }
            catch (Exception ex)
            {
                // In case of an exception, file an error message; the
                // system may be still functional, though.
                _log.error(ex);
                ex.printStackTrace();
            }
        }
//        _log.debug("Finished initializing all services (early).");
    }

    /**
     * Retrieves an instance of a service without triggering late
     * initialization.
     * 
     * Early initialization of a service can require access to service
     * properties.  The service must have its name and ServiceManager
     * set by then.  Therefore, before calling
     * Initable.initClass(Object), the class must be instantiated with
     * InitableBroker.getInitableInstance(), and
     * Service.setManagerBroker() and Service.setName() must be
     * called.  This calls for two - level accessing the services
     * instances.
     */
    private Service getServiceInstance(String name)
        throws ServiceException
    {
        Service service = (Service) _services.get(name);

        if (service == null)
        {
            String className = (String) _mapping.get(name);
            if (className == null)
            {
                throw new ServiceException(
                    "Unknown service [" + name + "] requested");
            }
            try
            {
                service = (Service)
                        Class.forName(className).newInstance();
            }
            // those two errors must be passed to the VM
            catch (ThreadDeath td)
            {
                throw td;
            }
            catch (OutOfMemoryError ome)
            {
                throw ome;
            }
            catch (Throwable t)
            {
                // Used to indicate error condition.
                String msg = null;

                if (t instanceof NoClassDefFoundError)
                {
                    msg = "A class referenced by " + className +
                            " is unavailable. Check your jars and classes.";
                }
                else if (t instanceof ClassNotFoundException)
                {
                    msg = "Class " + className +
                            " is unavailable. Check your jars and classes.";
                }
                else if (t instanceof ClassCastException)
                {
                    msg = "Class " + className +
                            " doesn't implement the Service interface";
                }
                else
                {
                    msg = "Failed to instantiate " + className;
                }

                throw new ServiceRuntimeException(msg, t);
            }
            service.setServiceManager(this);
            service.setName(name);
            _services.put(name, service);
        }

        return service;
    }
}
