/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan;


import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import solar.api.Event;
import solar.api.Request;
import solar.service.fap.DataSource;
import solar.service.fap.DataSink;
import solar.service.fap.InputPort;
import solar.service.fap.OutputPort;
import solar.service.fap.Pipe;
import solar.service.fap.FapException;


/**
 *  The wrapper class of graph nodes through which the resident sources and
 *  operators interact with other services of the local Planet.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: NodeHandler.java,v 1.1 2003/09/19 00:38:28 glchen Exp $
 */
class NodeHandler 
    implements DataSource, DataSink, Runnable, DanConstants
{
    protected List _inputs = new Vector();
    protected List _outputs = new Vector();

    protected GraphNode _gnode;
    protected boolean _pullOnly = false;
    protected boolean _pushOnly = false;

    protected Thread _runner;

    
    //=======================================================================
    // Constructors.
    //=======================================================================

    public NodeHandler(GraphNode gnode)
    {
        _gnode = gnode;
        _pullOnly = (gnode instanceof Source);
        _pushOnly = (gnode instanceof Application);
        _gnode.init(this);
    }


    //=======================================================================
    // Implementation of java.lang.Runnable interface.
    //=======================================================================

    public void run()
    {
        if (!(_gnode instanceof Source))
        {
            throw new IllegalStateException("Only source can be run as thread.");
        }
        try
        {
            Source source = (Source) _gnode;
            source.run();
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    

    //=======================================================================
    // Implementation of solar.service.fap.DataSource interface.
    //=======================================================================
    
    public void attachOutputPipe(Pipe pipe, OutputPort outport)
        throws FapException
    {
        _outputs.add(pipe); // TODO: check whether the port matches
    }

    public Pipe[] getOutputPipes()
    {
        return (Pipe[]) _outputs.toArray();
    }

    public OutputPort[] getOutputPorts()
    {
        return (OutputPort[]) _gnode.getOutputPorts().toArray();
    }
    
    public Object pullData(Object query, OutputPort outport)
        throws FapException
    {
        if (_pushOnly)
        {
            throw new IllegalStateException("Cannot pull from an application.");
        }
        try
        {
            return _gnode.handleQuery((Request)query, outport.getName());
        }
        catch (Exception ex)
        {
            throw new FapException(ex);
        }
    }


    //=======================================================================
    // Implementation of solar.service.fap.DataSink interface.
    //=======================================================================

    public void attachInputPipe(Pipe pipe, InputPort inport)
        throws FapException
    {
        _inputs.add(pipe);  // TODO: check whether the port matches
    }

    public Pipe[] getInputPipes()
    {
        return (Pipe[]) _inputs.toArray();
    }

    public InputPort[] getInputPorts()
    {
        return (InputPort[]) _gnode.getInputPorts().toArray();
    }
    
    public void pushData(Object data, InputPort inport)
        throws FapException
    {
        if (_pullOnly)
        {
            throw new IllegalStateException("Cannot push to a source.");
        }
        try
        {
            _gnode.handleEvent((Event)data, inport.getName());
        }
        catch (Exception ex)
        {
            throw new FapException(ex);
        }
    }

    
    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Start this graph node as a thread. */
    public void start()
    {
        if (!(_gnode instanceof Source))
        {
            throw new IllegalStateException("Only source can be run as thread.");
        }
        _runner = new Thread(this);
        _runner.start();
    }

    
    /** Publish an event to all its subscribers. */
    public void publish(Event evt)
        throws Exception
    {
        publish(evt, DAN_OUTPORT);
    }


    /** Publish an event to all its subscribers through a specified port. */
    public void publish(Event evt, String outport)
        throws Exception
    {
        Iterator itr = _outputs.iterator();
        while (itr.hasNext())
        {
            Pipe p = (Pipe) itr.next();
            p.pushData(evt);    // TODO: check whether the port matches
        }
    }
}
