/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan;


import solar.api.Id;
import solar.api.Guid;
import solar.api.Request;
import solar.api.Address;
import solar.service.dan.comm.PipeConnectRequest;
import solar.service.fap.DataSink;
import solar.service.fap.DataSource;
import solar.service.fap.InputPort;
import solar.service.fap.OutputPort;
import solar.service.fap.FapService;
import solar.service.fap.FapException;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;


/**
 *  A pipe connector is responsible to set up a pipe between a data source
 *  and a data sink.  It does not matter whether the two ends are both local
 *  or one is remote.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PipeConnector.java,v 1.3 2003/10/05 18:50:55 glchen Exp $
 */
class PipeConnector 
    implements RpcConstants, DanConstants
{
    protected DanService _dan;
    protected FapService _fap;
    protected RpcService _rpc;
    
    protected Address _laddr;
    protected NodeManager _nodeman;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public PipeConnector(DanService dan)
    {
        _dan = dan;
        _fap = dan.getFapService();
        _rpc = dan.getRpcService();
        _laddr = dan.getLocalAddress();
        _nodeman = dan.getNodeManager();
    }

    private Id makePipeId()
    {
        return Id.makeId("pipe-");
    }
    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Handle the pconnect request. */
    public Id pconnect(Request request)
        throws Exception
    {
        // get request arguments
        PipeConnectRequest pcr = (PipeConnectRequest) request;
        Guid sguid = pcr.getSourceGuid();
        Guid dguid = pcr.getSinkGuid();
        Address address = pcr.getAddress();
        Id rpid = pcr.getPipeId();
        int type = pcr.getPipeType();
        boolean client = pcr.isClient();

        // make a local pipe ID and set up the pipe
        Id lpid = makePipeId();
        if (sguid != null)  // this end is the source of pipe
        {
            _fap.pconnect(lpid, type, 
                (DataSource)_nodeman.getLocal(sguid), 
                new OutputPort(DAN_OUTPORT),
                address, rpid, 
                new InputPort(DAN_INPORT), 
                client);
        }
        else    // this end is the sink of pipe
        {
            _fap.pconnect(lpid, type, 
                address, rpid, 
                new OutputPort(DAN_OUTPORT),
                (DataSink)_nodeman.getLocal(dguid), 
                new InputPort(DAN_INPORT), 
                client);
        }
        return lpid;
    }

    /** Connects a local source to a remote sink. */
    public void pconnect(int type,
        Guid sguid, String outport,
        Address addr, Guid dguid, String inport,
        boolean client)             // whether remote party is a solar client
        throws DanException
    {
        // create a local pipe ID    
        Id lpid = makePipeId();
        // prepare RPC request
        Request req = new PipeConnectRequest(null, dguid,
            _laddr, lpid, type, client);
        try
        {
            // get the remote pipe ID from RPC call
            Id dpid = (Id) callRpc(addr, req);
            // set up the pipe
            _fap.pconnect(lpid, type, 
                (DataSource)_nodeman.getLocal(sguid),
                new OutputPort(outport),
                addr, dpid,
                new InputPort(inport),
                false);
        }
        catch (Exception ex)
        {
            throw new DanException("Failed to setup the pipe.", ex);
        }
    }


    /** Connects a remote source to a local sink. */
    public void pconnect(int type,
        Address addr, Guid sguid, String outport,
        Guid dguid, String inport,
        boolean client)             // whether remote party is a solar client
        throws DanException
    {
        // create a local pipe ID    
        Id lpid = makePipeId();
        // prepare RPC request
        Request req = new PipeConnectRequest(sguid, null, 
            _laddr, lpid, type, client);
        try
        {
            // get the remote pipe ID from RPC call
            Id dpid = (Id) callRpc(addr, req);
            // set up the pipe
            _fap.pconnect(lpid, type, 
                addr, dpid,
                new OutputPort(outport),
                (DataSink)_nodeman.getLocal(dguid),
                new InputPort(inport),
                false);
        }
        catch (Exception ex)
        {
            throw new DanException("Failed to setup the pipe.", ex);
        }
    }


    /** Connects a local source to a local sink. */
    public void pconnect(int type,
        Guid sguid, String outport,
        Guid dguid, String inport)
        throws DanException
    {
        try
        {
            _fap.pconnect(null, type,
                (DataSource)_nodeman.getLocal(sguid),
                new OutputPort(outport),
                (DataSink)_nodeman.getLocal(dguid),
                new InputPort(inport));
        }
        catch (FapException fe)
        {
            throw new DanException("Failed to setup the pipe.", fe);
        }
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    /** Invoke a RPC call with specified request. */
    private Object callRpc(Address addr, Request req)
        throws RpcException
    {
        return _rpc.invoke(addr, _dan.getName(), req, RPC_TIMEOUT);
    }
}
