/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan.lang;


import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.File;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import javax.xml.parsers.*;

import solar.api.ParsingException;
import solar.service.directory.PlainName;
import solar.service.fap.FapConstants;


/**
 *  An XML parser for FAP graph specification.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: XmlSpecParser.java,v 1.1 2003/09/25 15:38:53 glchen Exp $
 */
public class XmlSpecParser
	extends DefaultHandler
    implements SpecParser, FapConstants
{
	private static final String GRAPH        = "graph";
	private static final String FILTER       = "filter";
	private static final String NAME         = "name";
	private static final String CLASSNAME    = "classname";
	private static final String CODEBASE     = "codebase";
	private static final String ADDRESS      = "address";
	private static final String ARGUMENT     = "argument";
	private static final String VALUE        = "value";

    private static final String PIPE         = "pipe";
    private static final String TYPE         = "type";
    private static final String PUSH         = "push";
    private static final String PULL         = "pull";
    private static final String MIXED        = "mixed";
    private static final String SOURCE       = "source";
    private static final String SINK         = "sink";
    private static final String INPORT       = "inport";
    private static final String OUTPORT      = "outport";
	private static final String FLOW         = "flow";

	private static final SAXParserFactory _factory = 
		SAXParserFactory.newInstance();

    // current parsing state
    private GraphSpec _graph;
    private FilterSpec _filter;
    private PipeSpec _pipe;
    private int _pipeNum = 0;


    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public XmlSpecParser()
    {
    }


    //=======================================================================
    // Implementation of solar.service.dan.SpecParser interface.
    //=======================================================================

    public GraphSpec parse(File file)
        throws ParsingException
    {
		try
		{
			FileInputStream in = new FileInputStream(file);
			GraphSpec ret = parse(in);
			in.close();
			return ret;
		}
		catch (IOException ioe)
		{
			throw new ParsingException(ioe);
		}
    }

    public GraphSpec parse(InputStream in)
        throws ParsingException
    {
		return parse(new InputSource (in));
    }

    public GraphSpec parse(Reader reader)
        throws ParsingException
    {
		return parse(new InputSource(reader));
    }

    public GraphSpec parse(String xml)
        throws ParsingException
    {
		StringReader reader = new StringReader(xml);
		return parse(new InputSource(reader));
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

	public static GraphSpec parse(InputSource in)
		throws ParsingException
	{
		try
		{
			SAXParser sax = _factory.newSAXParser();
			XmlSpecParser parser = new XmlSpecParser();
			sax.parse(in, parser);
			return parser.getGraphSpec();
		}
        catch (SAXException saxe)
        {
			throw new ParsingException(saxe);
        }
		catch (ParserConfigurationException pce)
		{
			throw new ParsingException(pce);
		}
		catch (IOException ioe)
		{
			throw new ParsingException(ioe);
		}
	}


    //=======================================================================
    // Parsing methods.
    //=======================================================================

    /** Return recent parsed graph spec. */
    public GraphSpec getGraphSpec()
    {
        return _graph;
    }

	public void endDocument ()
	{
		// reset parsing state
        _filter = null;
	}

	public void startElement (String uri, String localName, 
        String name, Attributes attrs)
		throws SAXException
	{
		switch (name.charAt (0))
		{
		case 'g':
            assertEquals(name, GRAPH);
            startGraph(attrs);
			break;
		case 'f':
            assertEquals(name, FILTER);
            startFilter(attrs);
			break;
		case 'a':
            assertEquals(name, ARGUMENT);
            startArgument(attrs);
			break;
		case 'i':
            assertEquals(name, INPORT);
			startInport(attrs);
			break;
		case 'o':
            assertEquals(name, OUTPORT);
			startOutport(attrs);
			break;
		case 'p':
            assertEquals(name, PIPE);
			startPipe(attrs);
			break;

		default:
			assertEquals(name, "");	// will throw proper exception
		}
    }

	public void endElement(String uri, String localName, String name)
		throws SAXException
	{
		switch (name.charAt (0))
		{
		case 'f':
            assertEquals(name, FILTER);
            endFilter();
			break;
		}
	}

    private void startGraph(Attributes attrs)
    {
        _graph = new GraphSpec();
    }
    
    private void startFilter(Attributes attrs)
        throws SAXException
    {
        String name = attrs.getValue(NAME);
        if (name == null || !name.startsWith("$"))
            throw new SAXException(
                "Filter must have a name like $myop.");
        if (_graph == null || _graph.getFilter(name) != null)
            throw new SAXException(
                "Operator has already been defined.");
            
        String clazz = attrs.getValue(CLASSNAME);
        String codebase = attrs.getValue(CODEBASE);
        String address = attrs.getValue(ADDRESS);

        _filter = new FilterSpec();
        _filter.setName(name);
        _filter.setClassname(clazz);
        _filter.setCodebase(codebase);
        _filter.setAddress(address);
    }

    private void endFilter()
    {
        _graph.putFilter(_filter.getName(), _filter);
        _filter = null;
    }

    private void startArgument(Attributes attrs)
    {
        String name = attrs.getValue(NAME);
        String value = attrs.getValue(VALUE);
        _filter.addArgument(name, value);
    }

    private void startInport(Attributes attrs)
    {
        String name = attrs.getValue(NAME);
        _filter.addInputPort(name);
    }

    private void startOutport(Attributes attrs)
    {
        String name = attrs.getValue(NAME);
        _filter.addOutputPort(name);
    }

    private void startPipe(Attributes attrs)
        throws SAXException
    {
        _pipe = new PipeSpec();

        // get the pipe name, could be null and auto generated
        String name = attrs.getValue(NAME);
        name = (name==null) ? Integer.toString(_pipeNum++) : name;
        _pipe.setName(name);
        // get the pipe type, must be specified
        String type = attrs.getValue(TYPE);
        if (type == null || type.equals(PUSH))  // default type is PUSH
            _pipe.setType(FAP_PUSH);
        else if (type.equals(PULL))
            _pipe.setType(FAP_PULL);
        else if (type.equals(MIXED))
            _pipe.setType(FAP_MIXED);
        else
            throw new SAXException("Illegal pipe type " + type);
            
        // get the info for the pipe source
        String source = attrs.getValue(SOURCE);
        String outport = attrs.getValue(OUTPORT);
        if (source.startsWith("$"))
        {
            if (_graph.getFilter(source) == null)
               throw new SAXException("The source of the pipe is not "
                    + "a previously defined filter.");
            _pipe.setInput(source, outport);
            _graph.getFilter(source).addOutput(name);
        }
        else
        {
            try
            {
                _pipe.setInput(new PlainName(source), outport);
            }
            catch (ParsingException pe)
            {
                throw new SAXException(pe);
            }
        }
        
        // get the info for the pipe sink
        String sink = attrs.getValue(SINK);
        String inport = attrs.getValue(INPORT);
        if (_graph.getFilter(sink) == null)
        {
            throw new SAXException("The sink of the pipe is not "
                + "a previously defined filter.");
        }
        else
        {
            _pipe.setOutput(sink, inport);
            _graph.getFilter(sink).addInput(name);
        }

        // get the flow control policy of this pipe, could be null
        String flow = attrs.getValue(FLOW);
        _pipe.setFlow(flow);
        _graph.putPipe(name, _pipe);
    }

	private void assertEquals (String s1, String s2)
		throws SAXException
	{
		if (!(s1.equals (s2)))
			throw new SAXException("Illegal tag: " + s1);
	}


    //=======================================================================
    // Test.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        SpecParser p = new XmlSpecParser();
        p.parse(new File(args[0]));
    }
}
