/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.directory;


import java.util.List;
import java.util.Vector;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Attribute;


/**
 *  A flat directory is a hashtable with name strands as the keys.
 *  
 *  TODO: name expiration
 *  
 *  @author     Guanling Chen
 *  @version    $Id: FlatDirectory.java,v 1.5 2003/09/25 21:37:38 glchen Exp $
 */
public class FlatDirectory
    implements Directory
{
    class NameInfo
    {
        Set _records = new HashSet();
        Set _listeners = new HashSet();
    }

    protected Hashtable _directory;

    
    //=======================================================================
    //  Constructors.
    //=======================================================================

    /** Constructs an empty flat name space. */
    public FlatDirectory()
    {
        _directory = new Hashtable();
    }


    //=======================================================================
    //  Implementation of solar.service.directory.Directory interface.
    //=======================================================================
    
    public void addListener(Name pattern, 
        DirectoryListener listener)
    {
        // basically register the listener with every strand
        Iterator itr = pattern.getStrands().iterator();
        while (itr.hasNext())
        {
            String strand = (String) itr.next();
            NameInfo info = (NameInfo) _directory.get(strand);
            info = (info == null) ? new NameInfo() : info;
            info._listeners.add(listener);
            _directory.put(strand, info);
        }
    }

    public void removeListener(Name pattern, 
        DirectoryListener listener)
    {
        // only works by assuming one listener can only observe one pattern
        // so the removal does not affect others
        Iterator itr = pattern.getStrands().iterator();
        while (itr.hasNext())
        {
            String strand = (String) itr.next();
            NameInfo info = (NameInfo) _directory.get(strand);
            if (info != null)
                info._listeners.remove(listener);
        }
    }

    public List lookup(Name pattern)
    {
        List advList = new Vector(); // final return results

        // get the set of name records who name contains all strands
        // from the specified pattern
        Set ret = null;
        Iterator itr = pattern.getStrands().iterator();
        while (itr.hasNext())
        {
            String strand = (String) itr.next();
            NameInfo ni = (NameInfo) _directory.get(strand);
            if (ni == null)
                break;
            Set r = ni._records;
            if (ret == null)
                ret = r;
            else
                ret.retainAll(r);   // set intersection operation
        }
        ret = (ret == null) ? new HashSet() : ret;

        // now reconstruct full names from these name records
        itr = ret.iterator();
        while (itr.hasNext())
        {
            NameRecord r = (NameRecord) itr.next();
            Set s = new HashSet();  // contains all strands for single record
            Iterator kitr = _directory.keySet().iterator();
            while (kitr.hasNext())
            {
                String strand = (String) kitr.next();
                if (((NameInfo)_directory.get(strand))._records.contains(r))
                    s.add(strand);
            }
            Name n = mergeStrands(s);   // merge strands into a single name
            advList.add(new Advertisement(n, r));
        }
        return advList;
    }

    private Name mergeStrands(Set strands)
    {
        // merge all the strands into a tree of name nodes
        NameNode root = new NameNode("/");
        Iterator itr = strands.iterator();
        while (itr.hasNext())
        {
            String strand = (String) itr.next();
            String[] s = strand.split("=");
            NameNode cur = root;
            for (int i=0; i<s.length; i++)
            {
                String t = s[i];
                NameNode chi = cur.getChild(t);
                if (chi == null)
                    chi = cur.createChild(t);
                cur = chi;
            }
        }
        
        // finally, get all attributes from this directory tree and
        // constructs a full name
        return new PlainName(getAttributes(root));
    }

    private List getAttributes(NameNode node)
    {
        String l = node.getLabel();
        List ret = new Vector();
        Iterator itr = node.getChildren().iterator();
        while (itr.hasNext())
        {
            NameNode c = (NameNode) itr.next();
            if (c.isLeaf())
            {
                ret.add(new Attribute(l, c.getLabel()));
            }
            else
            {
                // non-leaf, recursively constructs the attribute
                if (node.isRoot())
                {
                    ret.addAll(getAttributes(c));
                }
                else
                {
                    Name r = new PlainName(getAttributes(c));
                    ret.add(new Attribute(l, r));
                }
            }
        }
        return ret;
    }

    public NameRecord advertise(Name name, NameRecord rec)
    {
        NameRecord ret = null;
        Set listeners = null;       // listeners to receive notification
        boolean existed = false;    // whether the name is already in space
        boolean migrated = false;   // whether the name has moved
        
        Iterator itr = name.getStrands().iterator();
        while (itr.hasNext())
        {
            String strand = (String) itr.next();
            NameInfo info = (NameInfo) _directory.get(strand);
            info = (info == null) ? new NameInfo() : info;
            // go through the records, check address change
            Iterator ritr = info._records.iterator();
            while (ritr.hasNext())
            {
                NameRecord r = (NameRecord) ritr.next();
                if (r.getId().equals(rec.getId()))
                {
                    existed = true;
                    ret = r;    // the previous record to return

                    if (!r.getAddress().equals(rec.getAddress()))
                        migrated = true;
                }
            }
            // add the record to name space
            info._records.add(rec);
            _directory.put(strand, info);
            // update the set of listeners
            if (listeners == null)
                listeners = info._listeners;
            else
                listeners.retainAll(info._listeners);
        }
        // sending out notification events to listeners
        itr = listeners.iterator();
        while (itr.hasNext())
        {
            DirectoryListener ob = (DirectoryListener) itr.next();
            if (!existed)
                ob.nameAdded(name, rec);
            if (migrated)
                ob.nameMigrated(name, ret, rec);
        }
        return ret;
    }

    public NameRecord unadvertise(Name name, NameRecord rec)
    {
        NameRecord ret = null;
        Set listeners = null;       // listeners to receive notification
        
        Iterator itr = name.getStrands().iterator();
        while (itr.hasNext())
        {
            String strand = (String) itr.next();
            NameInfo info = (NameInfo) _directory.get(strand);
            if (info != null)
            {
                // go through the records
                Iterator ritr = info._records.iterator();
                while (ritr.hasNext())
                {
                    NameRecord r = (NameRecord) ritr.next();
                    if (r.getId().equals(rec.getId()))
                    {
                        ret = r;
                        ritr.remove();
                    }
                }
                // update the set of listeners
                if (listeners == null)
                    listeners = info._listeners;
                else
                    listeners.retainAll(info._listeners);
            }
        }
        // sending out notification events to listeners
        itr = listeners.iterator();
        while (itr.hasNext())
        {
            DirectoryListener ob = (DirectoryListener) itr.next();
            ob.nameRemoved(name, ret);
        }
        return ret;
    }

    public NameRecord update(Name oldname, 
        NameRecord oldrec, Name newname, NameRecord newrec)
    {
        NameRecord ret = unadvertise(oldname, oldrec);
        advertise(newname, newrec);
        return ret;
    }
}
