/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dispatch;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Adu;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.service.Service;
import solar.service.BaseService;
import solar.service.ServiceManager;
import solar.service.ServiceException;
import solar.service.transport.Header;
import solar.service.transport.TransportListener;
import solar.service.transport.TransportService;
import solar.util.Queue;
import solar.util.QueueFactory;


/**
 *  An event disptaching service for Solar.  This dispatch service requires a
 *  transport service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SolarDispatchService.java,v 1.19 2003/10/05 18:50:56 glchen Exp $
 */
public class SolarDispatchService 
    extends BaseService
    implements DispatchService, Runnable
{
    class NetListener
        implements TransportListener
    {
        String _transport;
        NetListener(String transport)
        {
            _transport = transport;
        }
        public void dataReceived(Adu adu, Header header)
        {
            String service = header.getMultiplexId().toString();
            _queue.enqueue(
                new DispatchUnit(adu, service, _transport, header.getSource()));
        }
        public void connectionOpened(Address peer)
        {
            
        }
        public void connectionClosed(Address peer)
        {
            
        }
    }
    
    /** Logging facility. */
    protected static Logger _log = Logger.getLogger(SolarDispatchService.class);

    /** Service manager. */   
    protected ServiceManager _manager;
    
    /** A table of transport services. */
    protected Hashtable _transports = new Hashtable();

    /** Inbound event queue. */
    protected Queue _queue = QueueFactory.makeQueue();

    
    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // get the service manager
        _manager = getServiceManager();
        // get all transport services used by this dispatch service.
        Options config = getServiceManager().getServiceConfig(getName());
        String[] tps = config.getProperty("transport", "transport").split(" ");
        for (int i=0; i<tps.length; i++)
        {
            String tpname = tps[i];
            // get the transport service
            TransportService tp = (TransportService) getDependentService("fakefake", tpname);
            // create transport listener
            tp.addTransportListener(new NetListener(tpname));
            // put the transport in table
            _transports.put(tpname, tp);
        }
        // start dispatching thread
        new Thread(this, "dispatcher").start();
    }

    // should not be called since all data are disptached for other service
    public void handleDispatchUnit(DispatchUnit dunit)
        throws Exception
    {
        _log.warn("Dropping received data for dispatcher: " + dunit.getAdu());
    }

    // should not be called since all requests are disptached for other service
    public Object handleRequest(Request req)
        throws Exception
    {
        _log.warn("Dropping received request for dispatcher: " + req);
        throw new IllegalStateException("Not expected to handle request.");
    }


    //=======================================================================
    // Implementation of DispatchService interface.
    //=======================================================================

    public TransportService getTransportService(String name)
    {
        return (TransportService) _transports.get(name);
    }

    public Address getTransportAddress(String name)
    {
        TransportService tps = (TransportService) _transports.get(name);
        if (tps != null)
        {
            return tps.getLocalAddress();
        }
        else
        {
            throw new IllegalArgumentException(
                "Transport service " + name + " was not found.");
        }
    }

    public void dispatchEvent(Event evt, 
        Address dest, String service, String transport)
        throws Exception
    {
        TransportService ts = getTransportService(transport);
        if (ts == null)
        {
            throw new IllegalArgumentException(
                "Specified transport service not found.");
        }
        ts.send(evt, dest, new Id(service));
    }


    //=======================================================================
    // Implementation of java.lang.Runnable interface.
    //=======================================================================

    public void run()
    {
        // continuously dequeue the data and call service handler
        while(true)
        {
            try
            {
                DispatchUnit dunit = (DispatchUnit) _queue.dequeue();
                String dest = dunit.getServiceName();
                if (dest != null)
                {
                    Service srv = _manager.getService(dest);
                    if (srv != null)
                    {
                        srv.handleDispatchUnit(dunit);
                    }
                    else
                    {
                        _log.debug("Service handler not found, " 
                            + "dropping received data.");
                    }
                }
                else
                {
                    _log.debug("Service handler not specified, " 
                        + "dropping received data.");
                }
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                _log.error(ex);
            }
        }
    }

}
