/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.fap;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.List;
import java.util.Vector;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.fap.comm.PipePushEvent;
import solar.service.fap.comm.PipePullRequest;
import solar.service.rpc.RpcService;
import solar.service.transport.TransportService;


/**
 *  Implementation of Filter-and-Pipe (FAP) service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: FapClientService.java,v 1.13 2003/10/05 18:50:57 glchen Exp $
 */
public class FapClientService 
    extends BaseService
    implements FapService
{
    protected Logger _log = Logger.getLogger(FapClientService.class);

    protected DispatchService _dispatch;
    protected RpcService _rpc;
    protected TransportService _transport;
    protected Hashtable _pipes = new Hashtable();

    // a list of pipes that have either the source or sink on some Solar
    // clients, not on Planets. The reason for the different treatment
    // is that the routing/transport between client-planet and planet-planet
    // may be configurated differently
    protected List _clients = new Vector();

    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _rpc = (RpcService)
            getDependentService("edge_rpc", "edge_rpc");
        _transport = (TransportService) 
            getDependentService("edge_transport", "edge_transport");
    }

    public void handleEvent(Event evt)
        throws Exception
    {
        // special trick if we use Transport instead of ObjectTransport
        PipePushEvent ppe = (evt instanceof PipePushEvent) ? 
            (PipePushEvent) evt : new PipePushEvent(evt);
        Id pid = ppe.getPipeId();
        Object data = ppe.getData();
        getPipe(pid).pushData(data);
    }

    public Object handleRequest(Request req)
      throws Exception
    {
        PipePullRequest ppr = (PipePullRequest) req;
        Id pid = ppr.getPipeId();
        Object query = ppr.getQuery();
        return getPipe(pid).pullData(query);
    }

    
    //=======================================================================
    // Implementation of solar.service.fap.FapService interface.
    //=======================================================================
    
    public Id pconnect(Id pid, int type,
        DataSource source, OutputPort outport,
        DataSink sink, InputPort inport)
        throws FapException
    {
        // create an empty pipe
        pid = registerPipeId(pid);
        Pipe pipe = PipeFactory.makePipe(pid, type, this);

        // link the pipe with source and sink with local channel
        SourceChannel scc = new LocalSourceChannel(source, outport);
        SinkChannel skc = new LocalSinkChannel(sink, inport);
        pipe.attachSourceChannel(scc);
        pipe.attachSinkChannel(skc);
        source.attachOutputPipe(pipe, outport);
        sink.attachInputPipe(pipe, inport);

        // register the pipe
        registerPipe(pid, pipe);
        return pid;
    }

    public Id pconnect(Id pid, int type,
        Address raddr, Id rpid, OutputPort outport,
        DataSink sink, InputPort inport, boolean client)
        throws FapException
    {
        // check the pipe ID
        pid = registerPipeId(pid);
        // if the sink on a solar client instead of another planet
        if (client) _clients.add(pid);
        // create an empty pipe
        Pipe pipe = PipeFactory.makePipe(pid, type, this);

        // link the pipe with source and sink with remote channel
        SourceChannel scc = new RemoteSourceChannel(raddr, rpid, outport);
        SinkChannel skc = new LocalSinkChannel(sink, inport);
        pipe.attachSourceChannel(scc);
        pipe.attachSinkChannel(skc);
        sink.attachInputPipe(pipe, inport);

        // register the pipe
        registerPipe(pid, pipe);
        return pid;
    }

    public Id pconnect(Id pid, int type,
        DataSource source, OutputPort outport,
        Address raddr, Id rpid, InputPort inport, boolean client)
        throws FapException
    {
        // check the pipe ID
        pid = registerPipeId(pid);
        // if the sink on a solar client instead of another planet
        if (client) _clients.add(pid);
        // create an empty pipe
        Pipe pipe = PipeFactory.makePipe(pid, type, this);
        
        // link the pipe with source and sink with remote channel
        SourceChannel scc = new LocalSourceChannel(source, outport);
        SinkChannel skc = new RemoteSinkChannel(raddr, rpid, inport);
        pipe.attachSourceChannel(scc);
        pipe.attachSinkChannel(skc);
        source.attachOutputPipe(pipe, outport);

        // register the pipe        
        registerPipe(pid, pipe);
        return pid;
    }

    public Pipe getPipe(Id pid)
    {
        synchronized (_pipes)
        {
            return (Pipe) _pipes.get(pid);
        }
    }

    public DispatchService getDispatchService()
    {
        return _dispatch;
    }
    
    public RpcService getRpcService(Id pid)
    {
        // all the pipes on the client use the same (the only) 
        // RPC service
        return _rpc;
    }

    public TransportService getTransportService(Id pid)
    {
        // all the pipes on the client use the same (the only) 
        // transport service
        return _transport;
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    protected void registerPipe(Id pid, Pipe pipe)
    {
        synchronized (_pipes)
        {
            _pipes.put(pid, pipe);
        }
    }
    
    protected Id registerPipeId(Id pid)
        throws FapException
    {
        pid = (pid == null) ? makePipeId() : pid;
        synchronized (_pipes)
        {
            if (_pipes.get(pid) != null)
            {
                throw new FapException("Pipe ID already exists.");
            }
            return pid;
        }
    }
    
    protected synchronized Id makePipeId()
    {
        return Id.makeId("pipe-");
    }
}
