/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.multicast;


import java.io.Serializable;
import org.apache.log4j.Logger;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.api.Attribute;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.directory.Name;
import solar.service.directory.PlainName;
import solar.service.directory.NameRecord;
import solar.service.directory.Advertisement;
import solar.service.directory.DirectoryService;
import solar.service.directory.DirectoryException;
import solar.service.multicast.comm.*;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcException;
import solar.service.transport.TransportService;
import solar.util.StringUtil;


/**
 *  A naive multicast service does not attempt to build a dissemination
 *  tree and simply sends the message directly to every subscriber in
 *  best-effort. Notice groupId must be global unique for semantic 
 *  correctness.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: NaiveMulticastService.java,v 1.18 2003/10/05 18:50:58 glchen Exp $
 */
public class NaiveMulticastService 
    extends BaseService
    implements MulticastService, MulticastConstants
{
    /** The logger for this multicast service. */
    protected Logger _log = Logger.getLogger(NaiveMulticastService.class);
    
    protected DirectoryService _directory;
    protected RpcService _rpc;
    protected TransportService _transport;
    protected Address _laddr;

    protected Hashtable _groups = new Hashtable();
    protected Hashtable _members = new Hashtable();

    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================

    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top
        
        _directory = (DirectoryService) 
            getDependentService("directory", "directory");
        _rpc = (RpcService) 
            getDependentService("rpc", "rpc");
        _transport = (TransportService) 
            getDependentService("transport", "transport");
        _laddr = _transport.getLocalAddress();
    }

    public Object handleRequest(Request request)
      throws Exception
    {
        Object ret = null;
        if (request instanceof JoinRequest)
        {
            JoinRequest jr = (JoinRequest) request;
            Id groupId = jr.getGroupId();
            RemoteMulticastMember rmm = jr.getRemoteMulticastMember();
            rmm.setLogger(_log);
            rmm.setRpcService(_rpc, getName());
            doJoin(groupId, rmm, false);
        }
        else if (request instanceof LeaveRequest)
        {
            LeaveRequest lr = (LeaveRequest) request;
            Id groupId = lr.getGroupId();
            RemoteMulticastMember rmm = lr.getRemoteMulticastMember();
            rmm.setLogger(_log);
            rmm.setRpcService(_rpc, getName());
            doLeave(groupId, rmm, false);
        }
        else if (request instanceof MulticastRequest)
        {
            MulticastRequest mr = (MulticastRequest) request;
            Id groupId = mr.getGroupId();
            Event evt = mr.getMulticastEvent();
            doMulticast(groupId, evt, false);
        }
        else if (request instanceof ReceiveRequest)
        {
            ReceiveRequest rr = (ReceiveRequest) request;
            Id id = rr.getId();
            Event evt = rr.getEvent();
            MulticastMember mb = (MulticastMember) _members.get(id);
            if (mb != null) mb.handleMulticastEvent(evt);
            else throw new MulticastException("Requested MulticastMember not found.");
        }
        else
        {
            throw new RpcException(
                "Failed to handle multicast request " + request + ".");
        }
        return ret;
    }

    
    //=======================================================================
    // Implementations of solar.service.multicast.MulticastService interface.
    //=======================================================================

    public void create(Id groupId)
        throws MulticastException
    {
        MulticastGroup group = (MulticastGroup) _groups.get(groupId);
        if (group != null)
            throw new MulticastException(
            "MulticastGroup to be created already exists.");
        group = new MulticastGroup(groupId);
        
        // create and put a name for the group in the name space
        Name name = getNameByGroupId(groupId);
        try
        {
            _directory.advertise(name, groupId, null);
        }
        catch (DirectoryException ne)
        {
            throw new MulticastException("Failed to create group.", ne);
        }
        synchronized (_groups)
        {
            _groups.put(groupId, group);
        }
    }

    public void destroy(Id groupId)
        throws MulticastException
    {
        // only allow to destroy group created locally
        MulticastGroup group = (MulticastGroup) _groups.get(groupId);
        if (group == null)
            throw new MulticastException(
            "Failed to find the MulticastGroup to be destroyed.");

        // create the name to be removed from name space
        Name name = getNameByGroupId(groupId);
        try
        {
            _directory.unadvertise(name, groupId);
        }
        catch (DirectoryException ne)
        {
            throw new MulticastException("Failed to destroy group.", ne);
        }
        synchronized (_groups)
        {
            _groups.remove(groupId);
        }
    }

    public void join(Id groupId, MulticastMember member)
        throws MulticastException
    {
        doJoin(groupId, member, true);
    }

    public void leave(Id groupId, MulticastMember member)
        throws MulticastException
    {
        doLeave(groupId, member, true);
    }

    public void multicast(Id groupId, Event evt)
        throws MulticastException
    {
        doMulticast(groupId, evt, true);
    }


    //=======================================================================
    //  Private helper methods.
    //=======================================================================

    private Name getNameByGroupId(Id groupId)
    {
        List attr = new Vector();
        attr.add(new Attribute(GROUP_ID, groupId.toString()));
        Name name = new PlainName(attr);
        return name;
    }
    
    private void doJoin(Id groupId, MulticastMember member, boolean local)
        throws MulticastException
    {
        MulticastGroup group = (MulticastGroup) _groups.get(groupId);
        if (group != null)
        {
            group.addMember(member);
        }
        else
        {
            if (!local)
            {
                throw new MulticastException(
                    "Failed to find requested MulticastGroup.");
            }
            else
            {
                try
                {
                    // lookup mc group using name space
                    Name name = getNameByGroupId(groupId);
                    List ret = _directory.lookup(name);
                    if (ret.size() == 0)
                        throw new MulticastException(
                            "Failed to find requested MulticastGroup.");
                    // get the current address of the node listing the group
                    Advertisement adv = (Advertisement) ret.iterator().next();
                    NameRecord rec = adv.getNameRecord();
                    Address addr = rec.getAddress();
                    
                    // add mc member to locally maintained table
                    Id id = Id.makeId("multicast-");
                    _members.put(id, member);

                    // create a remote proxy for joining the group
                    RemoteMulticastMember rmm = new RemoteMulticastMember(id, _laddr);
                    Request req = new JoinRequest(groupId, rmm);
                    _rpc.invoke(addr, getName(), req, 3000);
                }
                catch (RpcException re)
                {
                    throw new MulticastException(
                        "Failed to join the group.", re);
                }
                catch (DirectoryException ne)
                {
                    throw new MulticastException(
                        "Failed to join the group.", ne);
                }
            }
        }
    }

    private void doLeave(Id groupId, MulticastMember member, boolean local)
        throws MulticastException
    {
        MulticastGroup group = (MulticastGroup) _groups.get(groupId);
        if (group != null)
        {
            group.removeMember(member);
        }
        else
        {
            if (!local)
            {
                throw new MulticastException(
                    "Failed to find requested MulticastGroup.");
            }
            else
            {
                try
                {
                    // lookup mc group using name space
                    Name name = getNameByGroupId(groupId);
                    List ret = _directory.lookup(name);
                    if (ret.size() == 0)
                        throw new MulticastException(
                            "Failed to find requested MulticastGroup.");
                    // get the current address of the node listing the group
                    Advertisement adv = (Advertisement) ret.iterator().next();
                    NameRecord rec = adv.getNameRecord();
                    Address addr = rec.getAddress();
                    
                    // remove mc member from locally maintained table
                    Id oid = null;
                    Iterator itr = _members.keySet().iterator();
                    while (itr.hasNext())
                    {
                        Id id = (Id) itr.next();
                        if (_members.get(id) == member) // reference comparison
                        {
                            oid = id;
                            break;
                        }
                    }
                    if (oid != null)
                        _members.remove(oid);
                    else
                        throw new MulticastException(
                            "The mc member to remove does not exist.");

                    // create a remote proxy for leaving the group
                    RemoteMulticastMember rmm = new RemoteMulticastMember(oid, _laddr);
                    Request req = new LeaveRequest(groupId, rmm);
                    _rpc.invoke(addr, getName(), req, 3000);
                }
                catch (RpcException re)
                {
                    throw new MulticastException(
                        "Failed to join the group.", re);
                }
                catch (DirectoryException ne)
                {
                    throw new MulticastException(
                        "Failed to join the group.", ne);
                }
            }
        }
    }

    private void doMulticast(Id groupId, Event evt, boolean local)
        throws MulticastException
    {
        MulticastGroup group = (MulticastGroup) _groups.get(groupId);
        if (group != null)
        {
            Iterator itr = group.getMembers().iterator();
            while (itr.hasNext())
            {
                MulticastMember m = (MulticastMember) itr.next();
                m.handleMulticastEvent(evt);
            }
        }
        else
        {
            if (!local)
            {
                throw new MulticastException(
                    "Failed to find requested MulticastGroup.");
            }
            else
            {
                try
                {
                    // lookup mc group using name space
                    Name name = getNameByGroupId(groupId);
                    List ret = _directory.lookup(name);
                    if (ret.size() == 0)
                        throw new MulticastException(
                            "Failed to find requested MulticastGroup.");
                    // get the current address of the node listing the group
                    Advertisement adv = (Advertisement) ret.iterator().next();
                    NameRecord rec = adv.getNameRecord();
                    Address addr = rec.getAddress();
                    
                    // request multicast events to all members
                    Request req = new MulticastRequest(groupId, evt);
                    _rpc.invoke(addr, getName(), req, 3000);
                }
                catch (RpcException re)
                {
                    throw new MulticastException(
                        "Failed to multicast events to the group.", re);
                }
                catch (DirectoryException ne)
                {
                    throw new MulticastException(
                        "Failed to multicast events to the group.", ne);
                }
            }
        }
    }
}
