/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack.proto;


import cern.jet.random.Uniform;
import cern.jet.random.engine.MersenneTwister;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;


/**
 *  This output buffer manager takes all currently non-empty buffers and
 *  dequeue one of them with uniform probability.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: FairQueueOBM.java,v 1.4 2003/10/13 13:06:57 glchen Exp $
 */
public class FairQueueOBM 
    implements OutputBufferManager
{
    protected Uniform _unigen;
    protected BaseProtocol.ProtocolState _state;
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public FairQueueOBM(BaseProtocol.ProtocolState state)
    {
        _state = state;
        _unigen = new Uniform(new MersenneTwister());
    }

    //=======================================================================
    // Implementation of BufferManager interface.
    //=======================================================================

    public void suspendSubscriber(Object subkey)
    {
        synchronized (_state)
        {
            // suspend all handles with specified key
            // note special treatment if subkey is type Id (not Address)
            // this is for subrootproto mobility support
            Iterator itr = _state._buffers.values().iterator();
            while (itr.hasNext()) {
                Hashtable handles = (Hashtable) itr.next();
                Iterator itr2 = handles.keySet().iterator();
                while (itr2.hasNext()) {
                    Object key = itr2.next();
                    BufferHandle handle = (BufferHandle) handles.get(key);
                    // once a handle is suspended, all its buffers are
                    // frozen and not illgible to output
                    if (key.equals(subkey) ||
                        (subkey instanceof Id 
                        && handle.hasSubscriber((Id)subkey))) {
                        if (!handle.isSuspended()) {
                            _state._size -= handle.size();
                            handle.suspend();
                        }
                    }
                }
            }
        }
    }

    public void resumeSubscriber(Object subkey)
    {
        synchronized (_state)
        {
            // resume all handles with specified key
            // note special treatment if subkey is type Id (not Address)
            // this is for subrootproto mobility support
            Iterator itr = _state._buffers.values().iterator();
            while (itr.hasNext()) {
                Hashtable handles = (Hashtable) itr.next();
                Iterator itr2 = handles.keySet().iterator();
                while (itr2.hasNext()) {
                    Object key = itr2.next();
                    BufferHandle handle = (BufferHandle) handles.get(key);
                    // once a handle is suspended, all its buffers are
                    // frozen and not illgible to output
                    if (key.equals(subkey) ||
                        (subkey instanceof Id 
                        && handle.hasSubscriber((Id)subkey))) {
                        if (handle.isSuspended()) {
                            handle.resume();
                            _state._size += handle.size();
                        }
                    }
                }
            }
            // now wait up the thread if events available
            if (_state._size > 0) {
                _state.notifyAll();
            }
        }
    }

    public BufferedEvent getEvent()
        throws InterruptedException
    {
        List ret = new Vector();
        synchronized (_state)
        {
            // block if no event in buffers
            if (_state._size == 0) {
                _state.wait();
            }
            // get a list of non-empty buffers
            Iterator itr = _state._buffers.values().iterator();
            while (itr.hasNext()) {
                Hashtable handles = (Hashtable) itr.next();
                Iterator itr2 = handles.values().iterator();
                while (itr2.hasNext()) {
                    BufferHandle handle = (BufferHandle) itr2.next();
                    if (handle.size() > 0) {
                        ret.add(handle);
                    }
                }
            }
            // randomly pick one buffer from the list to get event
            BufferHandle handle = 
                (BufferHandle) ret.get(_unigen.nextIntFromTo(0, ret.size()-1));
            int osize = handle.size();
            BufferedEvent be = handle.getBufferedEvent();
            int nsize = handle.size();
            _state._size -= osize - nsize;
            return be;
        }
    }
}
