/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pubsub;


import serp.util.Options;
import org.apache.log4j.Logger;

import solar.api.Id;
import solar.api.Guid;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.pubsub.comm.PubEvent;
import solar.service.pubsub.comm.SubscribeRequest;
import solar.service.pubsub.comm.PubRegisterRequest;
import solar.service.pubsub.comm.SubRegisterRequest;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;
import solar.service.transport.SocketAddress;


/**
 *  The PACK service that runs at the client side.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ClientPubSubService.java,v 1.5 2003/10/05 19:42:45 glchen Exp $
 */
public class ClientPubSubService 
    extends BaseService
    implements PubSubService, RpcConstants
{
    /** Logging facility. */
    protected static Logger _log = 
        Logger.getLogger(ClientPubSubService.class);

    protected Address _server;
    protected String _pubsubName;
    protected String _transportName;
    protected Address _transportAddress;

    protected RpcService _rpc;
    protected DispatchService _dispatch;

    protected EndPointManager _epman = new EndPointManager();
    
    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _pubsubName = getName();
        Options config = getServiceManager().getServiceConfig(_pubsubName);
        _transportName = config.getProperty("transport", "transport");
        String server = config.getProperty("server");
        if (server != null)
        {
            try
            {
                _server = new SocketAddress(server);
            }
            catch (Exception ex)
            {
                throw new ServiceException(ex);
            }
        }
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // retrieve needed services
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _transportAddress = 
            _dispatch.getTransportAddress(_transportName);
        _rpc = (RpcService) 
            getDependentService("rpc", "rpc");

    }

    public Object handleRequest(Request request)
        throws Exception
    {
        return null;
    }

    public void handleEvent(Event evt)
        throws Exception
    {
        // handle events for local subscribers
        PubEvent pe = (evt instanceof PubEvent) ? 
            (PubEvent) evt : new PubEvent(evt);
        Id subId = pe.getSubId();
        Subscriber subscriber = (Subscriber) _epman.getEndPoint(subId);
        subscriber.handleEvent(pe.getEvent());
    }

    //=======================================================================
    // Implementation of PackService interface.
    //=======================================================================
    
    public void addPublisher(Id pubId, Publisher publisher)
        throws PubSubException
    {
        assertNotRegistered(pubId);
        
        try
        {
            // create a request and ask proxy for a session ID
            PubRegisterRequest prr = new PubRegisterRequest(pubId);
            prr.setSourceAddress(_transportAddress);
            Id session = (Id) callRpc(prr);
            // register the end point and allocate a buffer
            _epman.putEndPoint(pubId, session, _transportAddress, publisher);
        }
        catch (RpcException re)
        {
            throw new PubSubException(re);
        }
    }

    public void addSubscriber(Id subId, Subscriber subscriber)
        throws PubSubException
    {
        assertNotRegistered(subId);
        
        try
        {
            // create a request and ask proxy for a session ID
            SubRegisterRequest srr = new SubRegisterRequest(subId);
            srr.setSourceAddress(_transportAddress);
            Id session = (Id) callRpc(srr);
            // register the end point and allocate a buffer
            _epman.putEndPoint(subId, session, _transportAddress, subscriber);
        }
        catch (RpcException re)
        {
            throw new PubSubException(re);
        }
    }
    
    public void subscribe(Id pubId, Id subId, Object extra)
        throws PubSubException
    {
        assertRegistered(subId);
        
        try
        {
            // create a request and ask proxy for subscription
            SubscribeRequest sr = new SubscribeRequest(pubId, subId, extra);
            sr.setSourceAddress(_transportAddress);
            callRpc(sr);
        }
        catch (RpcException re)
        {
            throw new PubSubException(re);
        }        
    }

    public void publish(Id pubId, Event evt)
        throws PubSubException
    {
        assertRegistered(pubId);
        
        Event pe = new PubEvent(pubId, null, evt);
        try
        {
            _dispatch.dispatchEvent(
                pe, _server, _pubsubName, _transportName);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            _log.error(ex);
        }
    }
    
    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private Guid makeGuid(Id id)
    {
        return new Guid(_transportAddress, id);
    }

    private void assertRegistered(Id id)
        throws PubSubException
    {
        if (!_epman.contains(id))
        {
            throw new PubSubException("The end point has not registered.");
        }
    }

    private void assertNotRegistered(Id id)
        throws PubSubException
    {
        if (_epman.contains(id))
        {
            throw new PubSubException("The end point has already registered.");
        }
    }
    
    private Object callRpc(Request request)
        throws RpcException
    {
        return _rpc.invoke(_server, _pubsubName, request, RPC_TIMEOUT);
    }
}
