/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pubsub;


import java.io.Serializable;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

import rice.pastry.NodeId;
import rice.pastry.NodeHandle;
import rice.pastry.PastryNode;
import rice.pastry.security.Credentials;
import rice.scribe.Topic;
import rice.scribe.Scribe;
import rice.scribe.IScribe;
import rice.scribe.IScribeApp;
import rice.scribe.messaging.ScribeMessage;

import solar.api.Id;
import solar.api.Event;
import solar.api.Address;
import solar.service.transport.PastryAddress;


/**
 *  This is the only Scribe application that multiplexes events among
 *  end-points.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ScribeManager.java,v 1.2 2003/10/05 18:51:00 glchen Exp $
 */
public class ScribeManager 
    implements IScribeApp
{
    interface ScribeListener
    {
        public void eventReceived(Event evt);
    }

    protected ScribeListener _listener;
    protected Scribe _scribe;
    protected Credentials _cred;


    //===================================================================
    // Constructors.
    //===================================================================
    
    public ScribeManager(ScribeListener listener, 
        PastryNode node, Credentials cred)
    {
        _listener = listener;
        _scribe = new Scribe(node, cred);
        _scribe.registerApp(this);
    }

    //===================================================================
    // Public interface methods.
    //===================================================================

    public void create(Id topic)
    {
        _scribe.create(_scribe.generateTopicId(topic.toString()), _cred);
    }

    public void join(Id topic)
    {
        _scribe.join(_scribe.generateTopicId(topic.toString()), this, _cred);
    }

    public void multicast(Id topic, Event evt)
    {
        _scribe.multicast(_scribe.generateTopicId(topic.toString()), evt, _cred);
    }

    //===================================================================
    // Implementation of rice.scribe.IScribeApp interface.
    //===================================================================

    /**
     * Invoked when the underlying Scribe substrate is ready. 
     * The Scribe substrate becomes ready as soon as the local Pastry node
     * on which it lies is ready. In order to get this upcall it is 
     * necessary that the IScribeApp registers itself to the Scribe 
     * substrate using the registerApp() method in IScribe interface.
     */
    public void scribeIsReady()
    {

    }

    /**
     * up-call invoked by scribe when a publish message is 'delivered'.
     */    
    public void receiveMessage(ScribeMessage msg)
    {
        _listener.eventReceived((Event)msg.getData());
    }

    /**
     * Called by Scribe before the node forwards a message to
     * its children in the multicast tree.
     */
    public void forwardHandler(ScribeMessage msg)
    {

    }

     /**
     * Invoked by Scribe after a child is added to or removed from one of
     * the node's children tables.
     */
    public void subscribeHandler(NodeId topicId, 
        NodeHandle child, boolean wasAdded, Serializable obj )
    {

    }
    
    /**
     * Invoked by Scribe just before the "repair" SUBSCRIBE message is sent
     * when a node suspects its parent is faulty.
     */
    public void faultHandler(ScribeMessage msg, NodeHandle faultyParent)
    {
        
    }

    /**
     * Invoked by underlying scribe when it receives a anycast Message, 
     * to check if there is a local application which can take care of 
     * this anycast message.
     * If application can service the request of anycast message, the
     * anycast message is not routed furthur(for DFS), else the anycast
     * message will be furthur routed following DFS.
     */
    public boolean anycastHandler(ScribeMessage msg)
    {
        return true;
    }

    /**
     * Upcall made by scribe to registered applications to
     * inform them that local node is now the root for some
     * topic, and hence can take appropriate action.
     */
    public void isNewRoot(NodeId topicId)
    {

    }

    /**
     * Upcall made by scribe to the registered applications to
     * inform them that a new parent is found for local node
     * in the topic tree.
     */
    public void newParent(NodeId topicId, 
        NodeHandle newParent, Serializable data)
    {

    }
}
