/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pubsub;


import serp.util.Options;
import org.apache.log4j.Logger;

import java.util.Set;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import solar.api.Id;
import solar.api.Guid;
import solar.api.Event;
import solar.api.Address;
import solar.api.Request;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.pubsub.comm.PubEvent;
import solar.service.pubsub.comm.SubscribeRequest;
import solar.service.pubsub.comm.PubRegisterRequest;
import solar.service.pubsub.comm.SubRegisterRequest;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcException;
import solar.service.rpc.RpcConstants;
import solar.service.transport.PastryTransport;
import solar.service.transport.TransportService;


/**
 *  The Pub/Sub service that runs at the server side and is built on top
 *  of Scribe.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ScribePubSubService.java,v 1.4 2003/10/05 19:42:45 glchen Exp $
 */
public class ScribePubSubService 
    extends BaseService
    implements PubSubService, ScribeManager.ScribeListener, RpcConstants
{
    /** Logging facility. */
    protected static Logger _log = 
        Logger.getLogger(ScribePubSubService.class);

    protected String _pubsubName;
    protected String _coreTransportName;
    protected String _edgeTransportName;
    protected Address _coreTransportAddress;

    protected RpcService _coreRpc;
    protected DispatchService _dispatch;
    protected TransportService _coreTransport;

    protected ScribeManager _scribeman;
    protected Hashtable _submap = new Hashtable();
    protected EndPointManager _epman = new EndPointManager();
    

    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _pubsubName = getName();
        Options config = getServiceManager().getServiceConfig(_pubsubName);
        _coreTransportName = config.getProperty("core_transport", "transport");
        _edgeTransportName = config.getProperty("edge_transport", "transport");

        initDependents();
    }

    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // retrieve needed services
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _coreRpc = (RpcService)
            getDependentService("core_rpc", "core_rpc");
        _coreTransport = (TransportService)
            getDependentService("core_transport", "core_transport");

        // get scribe manager
        PastryTransport pastry = 
            (PastryTransport) _coreTransport.getTransport();
        _coreTransportAddress = pastry.getLocalAddress();
        _scribeman = new ScribeManager(this,
            pastry.getPastryNode(), pastry.getCredentials());
    }
    
    public Object handleRequest(Request request)
        throws Exception
    {
        Object ret = null;
        if (request instanceof SubRegisterRequest)
        {
            SubRegisterRequest srr = (SubRegisterRequest) request;
            Id session = makeSession();
            doSubRegister(
                srr.getSubId(), session, srr.getSourceAddress(), null);
            ret = session;
        }
        else if (request instanceof PubRegisterRequest)
        {
            PubRegisterRequest prr = (PubRegisterRequest) request;
            Id session = makeSession();
            doPubRegister(
                prr.getPubId(), session, prr.getSourceAddress(), null);
            ret = session;
        }
        else if (request instanceof SubscribeRequest)
        {
            SubscribeRequest sr = (SubscribeRequest) request;
            Id pubId = sr.getPubId();
            Id subId = sr.getSubId();
            Object extra = sr.getExtra();
            doSubscribe(pubId, subId, extra);
        }
        else
        {
            throw new RpcException(
                "Failed to handle pub/sub request " + request + ".");
        }
        return ret;
    }

    public void handleEvent(Event evt)
        throws Exception
    {
        // special trick if we use Transport instead of ObjectTransport
        PubEvent pe = (evt instanceof PubEvent) ? 
            (PubEvent) evt : new PubEvent(evt);
        Id pubId = pe.getPubId();
        Id subId = pe.getSubId();
        if (subId == null)
        {
            // from the clients, process before relay to root
            assertRegistered(pubId);
            doPublish(pubId, pe.getEvent());
        }
        else
        {
            throw new IllegalArgumentException(
                "Unexpected event received: " + evt);
        }
    }

        
    //=======================================================================
    // Implementation of PubSubService interface.
    //=======================================================================
    
    public void addPublisher(Id pubId, Publisher publisher)
        throws PubSubException
    {
        assertNotRegistered(pubId);
        doPubRegister(pubId, makeSession(), null, publisher);
    }

    public void addSubscriber(Id subId, Subscriber subscriber)
        throws PubSubException
    {
        assertNotRegistered(subId);
        doSubRegister(subId, makeSession(), null, subscriber);
    }
    
    public void subscribe(Id pubId, Id subId, Object extra)
        throws PubSubException
    {
        assertRegistered(subId);
        doSubscribe(pubId, subId, extra);
    }

    public void publish(Id pubId, Event evt)
        throws PubSubException
    {
        assertRegistered(pubId);
        doPublish(pubId, evt);
    }

    //=======================================================================
    // Private helper methods.
    //=======================================================================

    protected void doPubRegister(Id pubId, Id session, 
        Address address, Publisher publisher)
        throws PubSubException
    {
        // create the Scribe topic
        _scribeman.create(pubId);
        // record this publisher as an end point
        _epman.putEndPoint(pubId, session, address, publisher);
    }

    protected void doSubRegister(Id subId, Id session, 
        Address address, Subscriber subscriber)
    {
        // record this subscriber as an end point
        _epman.putEndPoint(subId, session, address, subscriber);
    }
    
    protected void doSubscribe(Id pubId, Id subId, Object extra)
        throws PubSubException
    {
        // record the subId in the subscription list of pubId
        Set subs = (Set) _submap.get(pubId);
        if (subs == null) 
        {
            subs = new HashSet();
            _submap.put(pubId, subs);
        }
        subs.add(subId);
        // make local scribe join the topic's multicast tree
        _scribeman.join(pubId);
    }

    protected void doPublish(Id pubId, Event evt)
        throws PubSubException
    {
        PubEvent pevt = new PubEvent(pubId, pubId, evt);
        _scribeman.multicast(pubId, pevt);
    }

    //=======================================================================
    // Implementation of ScribeManager.ScribeListener interface.
    //=======================================================================

    public void eventReceived(Event evt)
    {
        _log.debug("received scribe event");
        PubEvent pevt = (PubEvent) evt;
        Id pubId = pevt.getPubId();
        Iterator itr = ((Set)_submap.get(pubId)).iterator();
        while (itr.hasNext())
        {
            try
            {
                Id subId = (Id) itr.next();
                Subscriber subscriber = (Subscriber) _epman.getEndPoint(subId);
                if (subscriber != null)
                {
                    subscriber.handleEvent(pevt.getEvent());
                }
                else
                {
                    pevt.setSubId(subId);
                    Address dest = _epman.getAddress(subId);
                    _dispatch.dispatchEvent(
                        evt, dest, _pubsubName, _edgeTransportName);
                }
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                _log.error(ex);
            }
        }
    }

    
    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private Id makeSession()
    {
        String stamp = Long.toString(System.currentTimeMillis());
        return Id.makeId("session-" + stamp + "-");
    }
    
    private Guid makeGuid(Id id)
    {
        return new Guid(_coreTransportAddress, id);
    }

    private void assertRegistered(Id id)
        throws PubSubException
    {
        if (!_epman.contains(id))
        {
            throw new PubSubException("The end point " + id + " is not registered.");
        }
    }

    private void assertNotRegistered(Id id)
        throws PubSubException
    {
        if (_epman.contains(id))
        {
            throw new PubSubException("The end point " + id + " has already registered.");
        }
    }

}
