/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import org.apache.log4j.Logger;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import solar.api.Id;
import solar.api.Adu;
import solar.api.Address;
import solar.api.AduFactory;
import solar.impl.EventFactory;
import solar.service.BaseService;
import solar.service.routing.RoutingService;
import solar.service.routing.RoutingException;
import solar.util.NetUtil;
import solar.util.StringUtil;


/**
 *  The base implementation of transport service. Subclass must use
 *  init() method to instantiate an actual transport.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: BaseTransportService.java,v 1.2 2003/10/05 18:51:01 glchen Exp $
 */
public class BaseTransportService
    extends BaseService
    implements TransportService, TransportListener
{
    /** The logger for Pastry transport service. */
    protected Logger _log = Logger.getLogger(BaseTransportService.class);
    
    /** The actual transport layer. */
    protected Transport _transport;

    /** The routing service has the routing/forwarding table. */
    protected RoutingService _routing;

    /** The factory parses received packets. */
    protected AduFactory _aduFactory = new EventFactory();;

    protected List _listeners = new Vector();

    //=======================================================================
    // Implementation of TransportListener interface.
    //=======================================================================

    public void addTransportListener(TransportListener listener)
    {
        if (_listeners == null)
        {
            _listeners = new Vector();
        }
        _listeners.add(listener);
    }

    public void removeTransportListener(TransportListener listener)
    {
        if (_listeners != null)
        {
            _listeners.remove(listener);
        }
    }

    public void dataReceived(Adu adu, Header header)
    {
        try
        {/*     ***** ROUTING disabled *****
            if (_routing != null)
            {
                Address dest = header.getDestination();

                boolean local = false;
                if (dest instanceof SocketAddress)
                {
                    String ip = NetUtil.getIpAddress((
                        (SocketAddress)dest).getInetAddress());
                    if (ip.equals("127.0.0.1"))
                    {
                        local = true;
                    }
                }

                if (local || dest.equals(getLocalAddress()))
                {
                    notifyDataReceived(adu, header);
                }
                else
                {
                    _log.debug("Dest: " + dest + ":" 
                        + " Local: " + getLocalAddress());
                    // relay the received data
                    doSend(adu, dest, header.getMultiplexId());
                    _log.debug("Relay data hopping through...");
                }
            }
            else
            {*/
                notifyDataReceived(adu, header);
            //}
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            _log.error(ex);
        }
    }

    public void connectionOpened(Address peer)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.connectionOpened(peer);
        }
    }

    public void connectionClosed(Address peer)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.connectionClosed(peer);
        }
    }

    protected void notifyDataReceived(Adu adu, Header header)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.dataReceived(adu, header);
        }
    }

    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void doSend(Adu adu, Address dest, Id multiplex)
        throws TransportException
    {
        Address rdest = dest;
        if (_routing != null)
        {
            try
            {
                rdest = _routing.getNextHop(dest);
            }
            catch (RoutingException re)
            {
                throw new TransportException(
                    "No route to reach " + dest + ".", re);
            }
        }
        _transport.send(adu, rdest, multiplex);
    }
    

    //=======================================================================
    // Implementation of solar.service.TransportService interface.
    //=======================================================================

    public Address getLocalAddress()
    {
        return _transport.getLocalAddress();
    }

    public Address getAddressByKey(String key)
        throws TransportException
    {
        return _transport.getAddressByKey(key);
    }

    public Transport getTransport()
    {
        return _transport;
    }
    
    public void send(Adu adu, Address dest)
        throws TransportException
    {
        send(adu, dest, null);
    }

    public void send(Adu adu, Address dest, Id multiplex)
        throws TransportException
    {
        assertInited();
        doSend(adu, dest, multiplex);
    }
}
