/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import org.apache.log4j.Logger;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.EOFException;
import java.net.Socket;
import java.net.SocketException;

import solar.api.Adu;
import solar.api.AduFactory;
import solar.impl.ByteArrayAdu;
import solar.util.NetUtil;


/**
 *  The connection between two TCP peers.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: TcpConnection.java,v 1.3 2003/10/12 00:58:39 glchen Exp $
 */
class TcpConnection
    implements Runnable
{
    /** The cookie identifying our protocol handshake. */
    private final byte[] COOKIE = {(byte)'j', (byte)'i', 
        (byte)'t', (byte)'t', (byte)'e', (byte)'r'};
            
    /** The local socket for the connection. */
    private Socket _sock;

    /** The peer address. */
    private SocketAddress _peer;
        
    /** The thread receiving messages from this connection. */
    private Thread _receiver;

    private DataInputStream _datain;
    private DataOutputStream _dataout;

    private ObjectInputStream _objin;
    private ObjectOutputStream _objout;

    private DataInput _in;
    private DataOutput _out;

    private boolean _serialize;
    private AduFactory _aduFactory;
    private TcpTransport _transport;


    //===================================================================
    // Constructor.
    //===================================================================
        
    /** Constructs a new connection from a local connected socket. */
    public TcpConnection(boolean serialize, Socket sock, SocketAddress peer)
        throws Exception
    {
        _serialize = serialize;
        _sock = sock;
        _peer = peer;
        if (_serialize)
        {
            // must follow strict opening order, otherwise hang
            if (peer == null)
            {
                _objout = new ObjectOutputStream(_sock.getOutputStream());
                _objin = new ObjectInputStream(_sock.getInputStream());
            }
            else
            {
                _objin = new ObjectInputStream(_sock.getInputStream());
                _objout = new ObjectOutputStream(_sock.getOutputStream());
            }
            _in = _objin;
            _out = _objout;
        }
        else
        {
            _datain = new DataInputStream(_sock.getInputStream());
            _dataout = new DataOutputStream(_sock.getOutputStream());
            _in = _datain;
            _out = _dataout;
        }
    }


    //===================================================================
    // Public interface methods.
    //===================================================================
        
    /** Initialize this connection. */
    public void init(TcpTransport transport)
    {
        _transport = transport;
        _aduFactory = transport.getAduFactory();
        _receiver = new Thread(this);
        _receiver.start();
    }
        
    /** Destory this connection. */
    public void destroy()
    {
        try
        {
            _sock.close();
            if (_serialize)
            {
                _objin.close();
                _objout.close();
            }
            else
            {
                _datain.close();
                _dataout.close();                
            }
        }
        catch (Exception ex)
        {
            _transport.getLogger().warn(ex);
        }
        _receiver = null;
    }

    /** Receive peer address from this connection. */
    public SocketAddress receivePeerAddress()
        throws Exception
    {
        byte[] cookie, buf;
        int len = 0;

        cookie = new byte[COOKIE.length];
        initCookie(cookie);
        
        _in.readFully(cookie, 0, cookie.length);
        if (!matchCookie(cookie))
            throw new SocketException("cookie does not match.");
            
        String peer_ip = _in.readUTF();
        int peer_port = _in.readInt();
        _peer = new SocketAddress(peer_ip, peer_port);
        return _peer;
    }


    /** Send local address to peer through this connection. */
    public void sendLocalAddress(SocketAddress addr)
        throws Exception
    {
        if (addr == null) 
        {
            throw new SocketException("Null local address to send.");
        }

        _out.write(COOKIE, 0, COOKIE.length);
        _out.writeUTF(NetUtil.getIpAddress(addr.getInetAddress()));
        _out.writeInt(_transport.getPort());
        if (_serialize)
        {
            _objout.flush();
        }
        else
        {
            _dataout.flush();
        }
    }

    /** Send a message through this connection. */
    public void send(byte[] data, Header header)
        throws Exception
    {
        send(new ByteArrayAdu(data), header);
    }

    /** Send a message through this connection. */
    public synchronized void send(Adu adu, Header header)
        throws Exception
    {
        Exception ex = null;
        try
        {
            doSend(adu, header);
        }
        catch (Exception e)
        {
            try
            {
                doSend(adu, header);
            }
            catch (Exception e2)
            {
                ex = e2;
            }
        }
        if (ex != null) throw ex;
    }
    


    //===================================================================
    // Private helper methods.
    //===================================================================

    private void doSend(Adu adu, Header header)
        throws Exception
    {
        if (_serialize)
        {
            _objout.writeObject(header);
            _objout.writeObject(adu);
            _objout.flush();
        }
        else
        {
            header.write(_dataout);
            byte[] data = adu.toBytes();
            _dataout.writeInt(data.length);
            _dataout.write(data, 0, data.length);
            _dataout.flush();
        }
    }

    private void initCookie(byte[] c) 
    {
        if(c != null)
            for(int i=0; i < c.length; i++)
                c[i]=0;
    }
    
    private boolean matchCookie(byte[] cookie) 
    {
        if(cookie == null || cookie.length != COOKIE.length) 
            return false;
        for(int i=0; i < COOKIE.length; i++)
            if(cookie[i] != cookie[i]) 
                return false;
        return true;
    }

        
    //===================================================================
    // Implementation of java.lang.Runnable interface.
    //===================================================================

    public void run()
    {
        while (_receiver != null)
        {
            try
            {
                Header header = new Header();
                Adu adu = null;
                if (_serialize)
                {
                    header = (Header) _objin.readObject();
                    adu = (Adu) _objin.readObject();
                }
                else
                {
                    header.read(_datain);
                    int datasize = _in.readInt();
                    byte[] data = new byte[datasize];
                    _datain.readFully(data, 0, datasize);
                    if (_aduFactory != null)
                    {
                        adu = _aduFactory.parse(data);
                    }
                    else
                    {
                        adu = new ByteArrayAdu(data);
                    }
                }
                _transport.notifyDataReceived(adu, header);
            }
            catch(OutOfMemoryError oome)
            {
                _transport.getLogger().warn(oome);
                break;
            }
            catch(EOFException eofe)  // peer closed connection
            {
                _transport.getLogger().info(eofe);
                break;
            }
            catch(IOException ioe) 
            {
                _transport.getLogger().info(ioe);
                break;
            }
            catch (Exception ex)
            {
                _transport.getLogger().warn(ex);
            }
        }
        destroy();
        _transport.notifyConnectionClosed(_peer);
    }
}
