/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import org.apache.log4j.Logger;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.net.Socket;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.BindException;
import java.util.Random;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import solar.api.Id;
import solar.api.Adu;
import solar.api.Address;
import solar.api.AduFactory;
import solar.impl.ByteArrayAdu;


/**
 *  The UDP transport layer, converting ADU to bytes for sending.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: UdpTransport.java,v 1.17 2003/10/12 03:56:35 glchen Exp $
 */
public class UdpTransport
    implements Transport, Runnable, TransportConstants
{
    /** Logging facility. */
    protected static Logger _log;

    /** Particular interface to bind to. */
    protected String _hostname;
    
    /** The listening server port. */
    protected int _port;

    /** The socket to listen to. */
    protected DatagramSocket _ssock;

    /** The receiver thread receiving new messages. */
    protected Thread _receiver;

    /** The list of transport listeners. */
    protected List _listeners = new Vector();
    
    /** The local address for UDP. */
    protected Address _laddr;

    /** The factory to parse received byte array into ADU. */
    protected AduFactory _aduFactory;

    
    //=======================================================================
    // Constructors.
    //=======================================================================

    public UdpTransport()
        throws Exception
    {
        this(null, 0);
    }

    public UdpTransport(int port)
        throws Exception
    {
        this(null, port);
    }

    
    public UdpTransport(String hostname, int port)
        throws Exception
    {
        _log = Logger.getLogger(UdpTransport.class);
        _hostname = hostname;
        _port = port;
        start();
    }


    //=======================================================================
    // Implementation of solar.service.transport.Transport interface.
    //=======================================================================

    public void setAduFactory(AduFactory factory)
    {
        _aduFactory = factory;
    }
    
    public Address getLocalAddress()
    {
        return _laddr;
    }

    public Address getAddressByKey(String key)
        throws TransportException
    {
        throw new TransportException("DHT function not supported.");
    }

    public void send(byte[] data, Address dest)
        throws TransportException
    {
        send(data, dest, null);
    }

    public void send(byte[] data, Address dest, Id multiplex)
        throws TransportException
    {
        try
        {
            // prepare the header
            Header header = new Header(multiplex);
            header.setSource(_laddr);
            header.setDestination(dest);
            // format the final data array
            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            DataOutputStream out = new DataOutputStream(bout);
            header.write(out);
            out.writeInt(data.length);
            out.write(data, 0, data.length);
            out.flush();
            data = bout.toByteArray();
            // construct the packet and send it
            SocketAddress rdest = (SocketAddress) dest;
            DatagramPacket dp = 
                new DatagramPacket(data, data.length, 
                rdest.getInetAddress(), rdest.getPort());
            _ssock.send(dp);
            bout.close();
            out.close();
        }
        catch (Exception ex)
        {
            throw new TransportException("Failed to send UDP packet.", ex);
        }        
    }

    public void send(Adu adu, Address dest)
        throws TransportException
    {
        send(adu, dest, null);
    }


    public void send(Adu adu, Address dest, Id multiplex)
        throws TransportException
    {
        send(adu.toBytes(), dest, multiplex);
    }
    
    public void addTransportListener(TransportListener listener)
    {
        _listeners.add(listener);
    }

    public void removeTransportListener(TransportListener listener)
    {
        _listeners.remove(listener);
    }

    public void stop()
    {
        if (_ssock != null)
        {
            try
            {
                _ssock.close();
                _ssock = null;
                _receiver.interrupt();
            }
            catch (Exception ex) {}
        }
        _laddr = null;
        _listeners.clear();
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    /**  Start the UDP transport. */
    private void start()
        throws Exception
    {
        Random ran = new Random(System.currentTimeMillis());
        if (_port > 0)
        {
            if (_hostname == null)
                _ssock = new DatagramSocket(_port);
            else
                _ssock = new DatagramSocket(
                    _port, InetAddress.getByName(_hostname));
        }
        else
        {
            _port = 1050 + ran.nextInt(500);
            while(true) 
            {
                try 
                {
                    if (_hostname == null)
                        _ssock = new DatagramSocket(_port);
                    else
                        _ssock = new DatagramSocket(
                            _port, InetAddress.getByName(_hostname));
                    break;
                }
                catch (BindException be)
                {
                    _port += ran.nextInt(100);
                }
            }
        }
        _receiver = new Thread(this);
        _receiver.start();
        _hostname = (_hostname == null) ? "localhost" : _hostname;
        _laddr = new SocketAddress(_hostname, _port);
        _log.info("UDP transport started at local address: " + _laddr);
    }


    /** Notify transport listeners about received message. */
    private void notifyDataReceived(Adu adu, Header header)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener l = (TransportListener) itr.next();
            l.dataReceived(adu, header);
        }
    }


    //=======================================================================
    //  Implementation of java.lang.Runnable interface.
    //=======================================================================

    public void run()
    {
        byte[] ret = new byte[TRANSPORT_UDP_MSS];
        DatagramPacket dp = new DatagramPacket(ret, TRANSPORT_UDP_MSS);
        while (true)
        {
            try
            {
                // receive packet and construct byte array
                _ssock.receive(dp);
                byte[] data = dp.getData();
                ByteArrayInputStream bin = new ByteArrayInputStream(data);
                DataInputStream in = new DataInputStream(bin);
                // read in the header information
                Header header = new Header();
                header.read(in);
                // read in the actual data
                int len = in.readInt();
                data = new byte[len];
                in.readFully(data);
                // notify the listener
                if (_aduFactory != null)
                {
                    notifyDataReceived(_aduFactory.parse(data), header);
                }
                else
                {
                    notifyDataReceived(new ByteArrayAdu(data), header);
                }
                bin.close();
                in.close();
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                _log.error(ex);
            }
        }
    }
}
