/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */
 
package solar.test;


import cern.jet.random.Uniform;
import cern.jet.random.Exponential;
import cern.jet.random.engine.MersenneTwister;
import edu.cornell.lassp.houle.RngPack.Ranlux;
import serp.util.Options;
import java.io.FileInputStream;
import java.net.InetAddress;

import solar.api.Id;
import solar.api.Event;
import solar.impl.EventFactory;
import solar.service.ServiceManager;
import solar.service.pack.client.PackClient;
import solar.service.pubsub.Publisher;


/**
 *  A standalone event publisher using PACK service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PackSender.java,v 1.18 2003/10/13 14:53:38 glchen Exp $
 */
public class PackSender 
    implements Publisher
{
    protected Uniform _unigen;

    protected int _batch;
    protected int _counter;
    protected Id _pubId;
    protected PackClient _client;
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public PackSender()
    {
        _unigen = new Uniform(new MersenneTwister());
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================
    
    public void run(PackClient client, Options opts)
        throws Exception
    {
        Id pubId = new Id(opts.getProperty("pubid", "pack-pub"));
        _pubId = pubId;
        _client = client;
        _client.attachPublisher(pubId, this);
        _batch = opts.getIntProperty("batch", 50);
        
        String tra = opts.getProperty("traffic", "uni");
        int interval = opts.getIntProperty("interval", 3000);
        if (tra.equals("uni")) 
        {
            goUniform(interval);
        }
        else if (tra.equals("exp")) 
        {
            double lambda = opts.getDoubleProperty("lambda", 0.1);
            goExponential(lambda, interval);
        }
        else 
        {
            throw new Exception("Unsupported traffic type: " + tra);
        }
    }

    private void goUniform(int interval)
        throws Exception
    {
        int cnt = 0;
        while(true)
        {
            long start = System.currentTimeMillis();
            publish(start);
            long sleep = interval - (System.currentTimeMillis() - start);
//            System.out.println(sleep + " *** " + evt);
            if (sleep > 0) {
                Thread.sleep(sleep);
            }
        }
    }

    private void goExponential(double lambda, int interval)
        throws Exception
    {
        Exponential exp = new Exponential(lambda, new Ranlux());
        int cnt = 0;
        boolean on = true;
        while (true)
        {
            long fliptime = getNextInterval(exp);
            System.out.println(fliptime);
            if (on)
            {
                long flipstart = System.currentTimeMillis();
                while (System.currentTimeMillis() - flipstart < fliptime)
                {
                    long start = System.currentTimeMillis();
                    publish(start);
                    long sleep = interval - (System.currentTimeMillis() - start);
//                    System.out.println(sleep + " *** " + evt);
                    if (sleep > 0) {
                        Thread.sleep(sleep);
                    }
                }
            }
            else
            {
                Thread.sleep(fliptime);
            }
            on = !on;
        }
    }

    private void publish(long timestamp)
        throws Exception
    {
        for (int i=0; i<_batch; i++) {
            Event evt = EventFactory.getPlainEvent(
                "[tstamp=" + Long.toString(timestamp) 
                + ", counter=" + Integer.toString(_counter++)
                + ", value=" + Integer.toString(_unigen.nextIntFromTo(0, 1000)) 
                + "]");
            _client.publish(_pubId, evt);
        }
    }

    private long getNextInterval(Exponential exp)
    {
        return (long) exp.nextDouble()*1000;
    }
    

    //=======================================================================
    // Main entrance.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        // load configuration options
        String cfg = opts.getProperty("cfg");
        if (cfg != null)
        {
            opts.load(new FileInputStream(cfg));
            opts.setFromCmdLine(args);
        }

        // get command line options
        String host = opts.getProperty("host", "localhost");
        int port = opts.getIntProperty("port", 5470);
        // create client facade object
        PackClient client = new PackClient(host, port, opts);
        new PackSender().run(client, opts);
    }
}
