/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package versus;


import serp.util.Options;
import java.lang.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.io.*;

import solar.api.Event;
import solar.api.Request;
import solar.api.Attribute;
import solar.impl.EventFactory;
import solar.service.dan.Source;
import solar.Client;


/**
 *  A location source that connects to the Versus badge server. 
 *  It parses the badge event stream, and produces location events. 
 *  It generates and publish an event every time it receives
 *  a new badge event from the badge server.
 *  
 *  <p>The published events contain following attributes:
 *  <ul>
 *  <li>badge_num   (badge number)
 *  <li>col_num     (collector number)
 *  <li>sen_num     (sensor number)
 *  <li>button      (whether button is pressed)
 *  <li>battery     (whether battery is still good)
 *  <li>motion      (whether the badge is moving)
 *  <li>tcount      (a rotating event counter for a badge)
 *  <li>timestamp   (when the event is received)
 *  </ul>
 *
 *  @author     Guanling Chen
 *  @author     Kazuhiro Minami
 *  @version    $Id: VersusSource.java,v 1.5 2003/09/19 00:38:32 glchen Exp $
 */

public class VersusSource
    extends Source 
{
    private static final byte SEPARATOR     = -69;  // int 187

    private static final int BADGE_HEADER   = -1;
    private static final int BADGE_ID       = 0;
    private static final int QUALIFIERS     = 1;
    private static final int COLLECTOR_NO   = 2;
    private static final int SENSOR_NO      = 3;
    private static final int FIELDS_NUM     = 4; // #fields to be parsed 
    private static final int START_INDEX    = 6; // The index following '[' 

    private String _badgeHost;      // address of badge server
    private int _badgePort;         // port number of badge server
    private boolean _standalone;    // whether this source act as standalone
    private DataInputStream _in;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public VersusSource()
    {
        _standalone = true;
    }
    
    public VersusSource(boolean standalone)
    {
        _standalone = standalone;
    }


    //=======================================================================
    // Implementation of solar.service.dan.Source interface.
    //=======================================================================

    public void run()
        throws Exception
    {
        String badge_num = null;  // Badge number
        String col_num = null;    // Collector number
        String sen_num = null;    // Sensor number
        String tcount = null;     // TCount
        boolean motion = false;   // Motion state(T:active, F:inactive)
        boolean battery = false;  // Battery stete(T:high, F:low)
        boolean button = false;   // Button state(T:pressed, F:unpressed)
        
        byte[] rec = new byte[50];
        while (true)
        {
            int len = _in.read(rec);
            int pos = 0;
            int turn = BADGE_HEADER;
            long timestamp = System.currentTimeMillis();
            for (int i = 0; i < len; i++)
            {
                boolean exit = false;
                byte val = rec[i];
                if (val != SEPARATOR && val != 13 && val != 10) continue;

                // ignore the previous partial data if the header shows up
                String data = new String(rec, pos, i-pos);
                if (data.equals("LiveBadges") && turn != BADGE_HEADER)
                    turn = BADGE_HEADER;
                    
                switch (turn)
                {
                case BADGE_HEADER:
                    if (!data.equals("LiveBadges")) exit = true;
                    break;
                case BADGE_ID:
                    badge_num = data;
                    //System.out.print(badge_num);
                    break;
                case QUALIFIERS:
                    //qualifier = data;
                    // extract tcount
                    String xcount = data.substring(data.length()-1);
                    tcount = Integer.valueOf(xcount, 16).toString();
                    // extract flags
                    try
                    {
                        int flags = Integer.parseInt(data.substring(
                            data.length()-2, data.length()-1));
                        motion = ( (flags & 1) == 1 );
                        battery = ( (flags & 2) == 2 );
                        button = ( (flags & 4) == 4 );
                    }
                    catch (NumberFormatException nfe)
                    {
                        exit = true;
                    }
                    //System.out.print(" " + tcount + " " + motion + battery + button);
                    break;
                case COLLECTOR_NO:
                    col_num = data;
                    //System.out.print(" " + col_num);
                    break;
                case SENSOR_NO:
                    sen_num = data;

                    List at = new Vector();
                    at.add(new Attribute("badge_num", badge_num));
                    at.add(new Attribute("col_num", col_num));
                    at.add(new Attribute("sen_num", sen_num));
                    at.add(new Attribute("tcount", tcount));
                    at.add(new Attribute("button", button ? "true" : "false"));
                    at.add(new Attribute("battery", battery ? "true" : "false"));
                    at.add(new Attribute("motion", motion ? "true" : "false"));
                    at.add(new Attribute("timestamp", Long.toString(timestamp)));
                    Event evt = EventFactory.getPlainEvent(at);
                    
                    // Publish the event
                    if (_standalone)
                        System.out.println(evt);
                    else
                        publish(evt);
                    //System.out.println(" " + sen_num);
                    break;
                default:
                    exit = true;
                    break;
                }
                pos = i+1;
                turn = (turn+1)% FIELDS_NUM;
                if (exit) break;
            }
        }
    }

    public Object handleQuery(Request query)
        throws Exception
    {
        return null;
    }

    
    //=======================================================================
    // Private helper methods.
    //=======================================================================
    
    /** Open the connection to the badge server, and creates the 
     	input stream. */
    private void init(String badgeHost, int badgePort) 
    {
        try 
        {
            // Connect to the badge server and create input stream
            Socket badgeSocket = new Socket(badgeHost, badgePort);
            _in = new DataInputStream(badgeSocket.getInputStream());
        }
        catch (UnknownHostException uhe)
        {
            uhe.printStackTrace(System.err);
            System.exit(1);
        }
        catch (IOException ioe) 
        {
            ioe.printStackTrace(System.err);
            System.exit(1);
        }
    }

    
    /**
     *  Read the configuration from command line and start the source
     *  as Versus client or a standalone program.  The recognized options
     *  are:
     *  
     *  <ul>
     *  <ui>-standalone: whether to run the source without Solar.<BR>
     *  <ui>-host: the host address of the connecting Planet.<BR>
     *  <ui>-port: the port number of the connecting Planet.<BR>
     *  <ui>-badgehost: the badge server address to connect.<BR>
     *  <ui>-badgeport: the badge server port to connect.<BR>
     *  </ul>
     *  
     *  The Solar name assigned for this source is:
     *  [source=versus, building=sudikoff, campus=dartmouth]
     */
    public static void main(String args[])
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        // read badge server address and port
        boolean standalone = opts.getBooleanProperty("standalone", false);
        String badgeHost = opts.getProperty("badgehost");
        int badgePort = opts.getIntProperty("badgeport");
        VersusSource src = new VersusSource(standalone);
        src.init(badgeHost, badgePort);

        // if standalone, run directly, otherwise, connect to a planet
        if (standalone)
        {
            src.run();
            return;
        }

        // read planet address and port number
        String host = opts.getProperty("host");
        int port = opts.getIntProperty("port");
        if (host == null)
        {
            throw new Exception("invalid argument for Planet.");
        }
        
       // create a static to for advertisement
        String name = "[source=versus, building=sudikoff, campus=dartmouth]";
                        
        Client client = new Client(host, port);        
        client.advertise(name, src);
    }   
}
