/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package versus;


import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;


/**
 *  A XML parser for Versus configuration file containing floorplan
 *  and badge mapping.
 *
 *  @author     Guanling Chen
 *  @version    $Id: XmlVersusParser.java,v 1.1 2003/07/21 14:23:05 glchen Exp $
 */
public class XmlVersusParser
    extends DefaultHandler
{
    private final SAXParserFactory _factory = 
        SAXParserFactory.newInstance();

    private Hashtable _floorplan; 
    private Hashtable _badges;

    private Hashtable _currentCollectorTable;


    //=======================================================================
    // Public interface methods.
    //=======================================================================
    
    /** Parse Versus configuration in an XML file. */
    public void parse(File file) 
        throws Exception
    {
        try
        {
            FileInputStream in = new FileInputStream(file);
            parse(in);
            in.close();
        }
        catch (IOException ioe) 
        {
            throw new Exception(ioe);
        }
    }


    /** Parse Versus configuration in an XML stream. */
    public void parse(InputStream in) 
        throws Exception
    {
        parse(new InputSource(in));
    }


    /**  Parse Versus configuration in an XML input source. */
    public void parse(InputSource in) 
        throws Exception
    {
        try 
        {
            SAXParser sax = _factory.newSAXParser();
            sax.parse(in, this);
        }
        catch (ParserConfigurationException pce) 
        {
            throw new SAXException(pce);
        }
        catch (IOException ioe) 
        {
            throw new SAXException(ioe);
        }
    }


    /** Return the parsed floorplan. */
    public Hashtable getFloorplan()
    {
        return _floorplan;
    }


    /** Return the parsed badge mapping. */
    public Hashtable getBadges()
    {
        return _badges;
    }


    //=======================================================================
    // Implementation of default handler interface.
    //=======================================================================

    /** Triggered whenever a new XML tag is encountered. */
    public void startElement(String uri, String localName, 
        String name, Attributes attrs) 
        throws SAXException
    {
        switch(name.charAt(0)) 
        {
        case 'v':
            if (name.equals("versus"))
                startVersus(attrs);
            break;
        case 'c':
            if (name.equals("collector"))
                startCollector(attrs);
            break;
        case 's':
            if (name.equals("sensor"))
                startSensor(attrs);
            break;
        case 'b':
            if (name.equals("badge"))
                startBadge(attrs);
            break;
        default:
            throw new SAXException("Illegal tag: " + name);
        }
    }

    /** Triggered whenever a XML tag has been parsed. */
    public void endElement(String uri, String localName, String name)
    {
        switch(name.charAt(0)) 
        {
        case 'c':
            if (name.equals("collector"))
                _currentCollectorTable = null;
            break;
        }
    }

    private void startVersus(Attributes attrs)
    {
        _floorplan = new Hashtable();
        _badges = new Hashtable();
    }

    private void startCollector(Attributes attrs)
    {
        String id = attrs.getValue("id");
        _currentCollectorTable = (Hashtable)_floorplan.get(id);
        if (_currentCollectorTable == null) 
        {
            _currentCollectorTable = new Hashtable();
            _floorplan.put(id, _currentCollectorTable);
        }
    }

    private void startSensor(Attributes attrs)
    {
        String id = attrs.getValue("id");
        String zone = attrs.getValue("zone");

        if (_currentCollectorTable != null)
            _currentCollectorTable.put(id, zone);
    }

    private void startBadge(Attributes attrs)
    {
        String id = attrs.getValue("id");
        String name = attrs.getValue("name");

        _badges.put(id, name);
    }

}
